"""This dummy dataset is intended for testing and debugging purposes, not for
benchmarking. If your approach can fit this dataset it means that it is able to learn
how to estimate RUL. It does not mean it is good at it. """

from typing import Tuple, List, Optional, Union

import numpy as np
from sklearn import preprocessing  # type: ignore

from rul_datasets import utils
from rul_datasets.reader import AbstractReader, scaling


class DummyReader(AbstractReader):
    """
    This reader represents a simple, small dummy dataset that can be uses to test or
    debug RUL estimation approaches. It contains ten runs for each split with a
    single feature which makes it easy to hold in memory even on low-end computers.
    The dataset is so simple that it can be sufficiently fit by a three-layer
    perceptron in less than 50 epochs.

    Each run is randomly generated by sampling a run length between 90 and 110 time
    steps and creating a piece-wise linear RUL function `y(t)` with a maximum value of
    `max_rul`. The feature `x(t)` is then calculated as:

    ```python
    x(t) = exp(-0.05 * y(t) + N(offset, 0.01)) + N(0, noise_factor)
    ```

    where `N(loc, scale)` is a function drawing a sample from a normal distribution
    with a mean of `loc` and a standard deviation of `scale`. The `dev`, `val` and
    `test` splits are all generated the same way with a different fixed random seed.
    This makes generating the dataset deterministic.

    The dummy dataset contains two sub-datasets. The first has uses an `offset` of
    0.5 and a `noise_factor` of 0.01. The second uses an `offset` of 0.75 and a
    `noise_factor` of 0.02. Both use a default window size of 10 and are min-max
    scaled between -1 and 1 with a scaler fitted on the `dev` split.

    Examples:
        >>> import rul_datasets
        >>> fd1 = rul_datasets.reader.DummyReader(fd=1)
        >>> features, labels = fd1.load_split("dev")
        >>> features[0].shape
        (81, 10, 1)
    """

    _FDS = [1, 2]

    _DEFAULT_WINDOW_SIZE = 10
    _NOISE_FACTOR = {1: 0.01, 2: 0.02}
    _OFFSET = {1: 0.5, 2: 0.75}
    _SPLIT_SEED = {"dev": 42, "val": 1337, "test": 101}

    def __init__(
        self,
        fd: int,
        window_size: Optional[int] = None,
        max_rul: Optional[int] = 50,
        percent_broken: Optional[float] = None,
        percent_fail_runs: Optional[Union[float, List[int]]] = None,
        truncate_val: bool = False,
        truncate_degraded_only: bool = False,
    ):
        """
        Create a new dummy reader for one of the two sub-datasets. The maximum RUL
        value is set to 50 by default. Please be aware that changing this value will
        lead to different features, too, as they are calculated based on the RUL
        values.

        For more information about using readers, refer to the [reader]
        [rul_datasets.reader] module page.

        Args:
            fd: Index of the selected sub-dataset
            window_size: Size of the sliding window. Default defined per sub-dataset.
            max_rul: Maximum RUL value of targets.
            percent_broken: The maximum relative degradation per time series.
            percent_fail_runs: The percentage or index list of available time series.
            truncate_val: Truncate the validation data with `percent_broken`, too.
            truncate_degraded_only: Only truncate the degraded part of the data
                                    (< max RUL).
        """
        super(DummyReader, self).__init__(
            fd,
            window_size,
            max_rul,
            percent_broken,
            percent_fail_runs,
            truncate_val,
            truncate_degraded_only,
        )

        features, _ = self._generate_split("dev")
        scaler = preprocessing.MinMaxScaler(feature_range=(-1, 1))
        self.scaler = scaling.fit_scaler(features, scaler)

    @property
    def dataset_name(self) -> str:
        return "xjtu-sy"

    @property
    def fds(self) -> List[int]:
        """Indices of available sub-datasets."""
        return self._FDS

    def default_window_size(self, fd: int) -> int:
        return self._DEFAULT_WINDOW_SIZE

    def prepare_data(self) -> None:
        """This function has no effect as there is nothing to prepare."""
        pass

    def load_complete_split(
        self, split: str, alias: str
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        features, targets = self._generate_split(split, alias)
        features = scaling.scale_features(features, self.scaler)

        return features, targets

    def _generate_split(
        self, split: str, alias: Optional[str] = None
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        alias = split if alias is None else alias
        features = []
        targets = []
        rng = np.random.default_rng(self._SPLIT_SEED[split])
        for i in range(10):
            t = self._generate_targets(rng)
            f = self._generate_features(rng, t)
            f = utils.extract_windows(f, self.window_size)
            t = t[: -(self.window_size - 1)]
            features.append(f)
            targets.append(t)
        if alias == "test":
            features, targets = self._truncate_test_split(rng, features, targets)

        return features, targets

    def _generate_targets(self, rng):
        length = rng.integers(90, 110)
        t = np.clip(np.arange(length, 1, -1), a_min=0, a_max=self.max_rul)
        t = t.astype(float)

        return t

    def _generate_features(self, rng, targets):
        steady = -0.05 * targets + self._OFFSET[self.fd] + rng.normal() * 0.01
        noise = rng.normal(size=targets.shape) * self._NOISE_FACTOR[self.fd]
        f = np.exp(steady) + noise

        return f[:, None]

    def _truncate_test_split(self, rng, features, targets):
        """Extract a single window from a random position of the time series."""
        for i in range(len(features)):
            run_len = len(features[i])
            cutoff = rng.integers(run_len // 2, run_len - 1)
            features[i] = features[i][None, cutoff]
            targets[i] = targets[i][None, cutoff]

        return features, targets



"""The NASA CMAPSS Turbofan Degradation dataset is a collection of simulated
degradation experiments on jet engines. It contains four sub-datasets named FD1, FD2,
FD3 and FD4 which differ in operation conditions and possible failure types."""

import os
import tempfile
import warnings
import zipfile
from typing import Union, List, Tuple, Dict, Optional

import numpy as np
from sklearn import preprocessing as scalers  # type: ignore

from rul_datasets.reader import scaling
from rul_datasets.reader.data_root import get_data_root
from rul_datasets.reader.abstract import AbstractReader
from rul_datasets import utils


CMAPSS_URL = "https://kr0k0tsch.de/rul-datasets/CMAPSSData.zip"


class CmapssReader(AbstractReader):
    """
    This reader represents the NASA CMAPSS Turbofan Degradation dataset. Each of its
    four sub-datasets contains a training and a test split. Upon first usage,
    the training split will be further divided into a development and a validation
    split. 20% of the original training split is reserved for validation.

    The features are provided as sliding windows over each time series in the
    dataset. The label of a window is the label of its last time step. The RUL labels
    are capped by a maximum value. The original data contains 24 channels per time
    step. Following the literature, we omit the constant channels and operation
    condition channels by default. Therefore, the default channel indices are 4, 5,
    6, 9, 10, 11, 13, 14, 15, 16, 17, 19, 22 and 23.

    The features are min-max scaled between -1 and 1. The scaler is fitted on the
    development data only.

    Examples:
        Default channels
        >>> import rul_datasets
        >>> fd1 = rul_datasets.reader.CmapssReader(fd=1, window_size=30)
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> features[0].shape
        (163, 30, 14)

        Custom channels
        >>> import rul_datasets
        >>> fd1 = rul_datasets.reader.CmapssReader(fd=1, feature_select=[1, 2, 3])
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> features[0].shape
        (163, 30, 3)
    """

    _FMT: str = (
        "%d %d %.4f %.4f %.1f %.2f %.2f %.2f %.2f %.2f %.2f %.2f %.2f "
        "%.2f %.2f %.2f %.2f %.2f %.2f %.4f %.2f %d %d %.2f %.2f %.4f"
    )
    _TRAIN_PERCENTAGE: float = 0.8
    _WINDOW_SIZES: Dict[int, int] = {1: 30, 2: 20, 3: 30, 4: 15}
    _DEFAULT_CHANNELS: List[int] = [4, 5, 6, 9, 10, 11, 13, 14, 15, 16, 17, 19, 22, 23]
    _NUM_TRAIN_RUNS: Dict[int, int] = {1: 80, 2: 208, 3: 80, 4: 199}
    _CONDITION_BOUNDARIES: List[Tuple[float, float]] = [
        (-0.009, 0.009),  # Different from paper to include FD001 and FD003
        (9.998, 10.008),
        (19.998, 20.008),
        (24.998, 25.008),
        (34.998, 35.008),
        (41.998, 42.008),
    ]
    _CONDITION_COLUMN: int = 0
    _CMAPSS_ROOT: str = os.path.join(get_data_root(), "CMAPSS")

    def __init__(
        self,
        fd: int,
        window_size: Optional[int] = None,
        max_rul: Optional[int] = 125,
        percent_broken: Optional[float] = None,
        percent_fail_runs: Optional[Union[float, List[int]]] = None,
        feature_select: Optional[List[int]] = None,
        truncate_val: bool = False,
        operation_condition_aware_scaling: bool = False,
        truncate_degraded_only: bool = False,
    ) -> None:
        """
        Create a new CMAPSS reader for one of the sub-datasets. The maximum RUL value
        is set to 125 by default. The 14 feature channels selected by default can be
        overridden by passing a list of channel indices to `feature_select`. The
        default window size is defined per sub-dataset as the minimum time series
        length in the test set.

        The data can be scaled separately for each operation condition, as done by
        Ragab et al. This only affects FD002 and FD004 due to them having multiple
        operation conditions.

        For more information about using readers refer to the [reader]
        [rul_datasets.reader] module page.

        Args:
            fd: Index of the selected sub-dataset
            window_size: Size of the sliding window. Default defined per sub-dataset.
            max_rul: Maximum RUL value of targets.
            percent_broken: The maximum relative degradation per time series.
            percent_fail_runs: The percentage or index list of available time series.
            feature_select: The index list of selected feature channels.
            truncate_val: Truncate the validation data with `percent_broken`, too.
            operation_condition_aware_scaling: Scale data separatly for each
                                               operation condition.
            truncate_degraded_only: Only truncate the degraded part of the data
                                    (< max RUL).
        """
        super().__init__(
            fd,
            window_size,
            max_rul,
            percent_broken,
            percent_fail_runs,
            truncate_val,
            truncate_degraded_only,
        )
        # Select features according to https://doi.org/10.1016/j.ress.2017.11.021
        if feature_select is None:
            feature_select = self._DEFAULT_CHANNELS
        self.feature_select = feature_select
        self.operation_condition_aware_scaling = operation_condition_aware_scaling

    @property
    def dataset_name(self) -> str:
        return "cmapss"

    @property
    def fds(self) -> List[int]:
        """Indices of available sub-datasets."""
        return list(self._WINDOW_SIZES)

    def default_window_size(self, fd: int) -> int:
        return self._WINDOW_SIZES[fd]

    def prepare_data(self) -> None:
        """
        Prepare the CMAPSS dataset. This function needs to be called before using the
        dataset for the first time.

        The dataset is downloaded from a custom mirror and extracted into the data
        root directory. The training data is then split into development and
        validation set. Afterwards, a scaler is fit on the development features.
        Previously completed steps are skipped.
        """
        if not os.path.exists(self._CMAPSS_ROOT):
            _download_cmapss(get_data_root())
        # Check if training data was already split
        if not os.path.exists(self._get_feature_path("dev")):
            warnings.warn(
                f"Training data for FD{self.fd:03d} not "
                f"yet split into dev and val. Splitting now."
            )
            self._split_fd_train(self._get_feature_path("train"))
        if not os.path.exists(self._get_scaler_path()):
            self._prepare_scaler()

    def _prepare_scaler(self) -> None:
        dev_features, ops_cond = self._load_features(self._get_feature_path("dev"))
        dev_features, _ = self._split_time_steps_from_features(dev_features)
        scaler = self._fit_scaler(dev_features, ops_cond)
        scaling.save_scaler(scaler, self._get_scaler_path())

    def _fit_scaler(self, features, operation_conditions):
        scaler = scalers.MinMaxScaler(feature_range=(-1, 1))
        if self.operation_condition_aware_scaling:
            scaler = scaling.OperationConditionAwareScaler(
                scaler, self._CONDITION_BOUNDARIES
            )
            operation_conditions = [
                c[:, self._CONDITION_COLUMN] for c in operation_conditions
            ]
            scaler = scaling.fit_scaler(features, scaler, operation_conditions)
        else:
            scaler = scaling.fit_scaler(features, scaler)

        return scaler

    def _split_fd_train(self, train_path: str) -> None:
        train_data = np.loadtxt(train_path)

        # Split into runs
        _, samples_per_run = np.unique(train_data[:, 0], return_counts=True)
        split_indices = np.cumsum(samples_per_run)[:-1]
        split_train_data = np.split(train_data, split_indices, axis=0)

        split_idx = int(len(split_train_data) * self._TRAIN_PERCENTAGE)
        dev_data = np.concatenate(split_train_data[:split_idx])
        val_data = np.concatenate(split_train_data[split_idx:])

        data_root, train_file = os.path.split(train_path)
        dev_file = train_file.replace("train_", "dev_")
        dev_file = os.path.join(data_root, dev_file)
        np.savetxt(dev_file, dev_data, fmt=self._FMT)  # type: ignore
        val_file = train_file.replace("train_", "val_")
        val_file = os.path.join(data_root, val_file)
        np.savetxt(val_file, val_data, fmt=self._FMT)  # type: ignore

    def _get_scaler_path(self) -> str:
        return os.path.join(self._CMAPSS_ROOT, self._get_scaler_name())

    def _get_scaler_name(self) -> str:
        ops_aware = "_ops_aware" if self.operation_condition_aware_scaling else ""
        name = f"FD{self.fd:03d}_scaler_{self.feature_select}{ops_aware}.pkl"

        return name

    def _get_feature_path(self, split: str) -> str:
        return os.path.join(self._CMAPSS_ROOT, self._get_feature_name(split))

    def _get_feature_name(self, split: str) -> str:
        return f"{split}_FD{self.fd:03d}.txt"

    def load_complete_split(
        self, split: str, alias: str
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        file_path = self._get_feature_path(split)
        features, operation_conditions = self._load_features(file_path)
        features, time_steps = self._split_time_steps_from_features(features)
        features = self._scale_features(features, operation_conditions)

        if split in ["dev", "val"]:
            targets = self._generate_targets(time_steps)
        elif split == "test":
            targets = self._load_targets(features)
        else:
            raise ValueError(f"Unknown split {split}.")

        features, targets = self._pad_data(features, targets)

        if alias in ["dev", "val"]:
            features, targets = self._window_data(features, targets)
        elif alias == "test":
            features, targets = self._crop_data(features, targets)
        else:
            raise ValueError(f"Unknown alias {alias}.")

        return features, targets

    def _load_features(
        self, file_path: str
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        raw_features = np.loadtxt(file_path)

        feature_idx = [0, 1] + [idx + 2 for idx in self.feature_select]
        operation_conditions = raw_features[:, [2, 3, 4]]
        raw_features = raw_features[:, feature_idx]

        # Split into runs
        _, samples_per_run = np.unique(raw_features[:, 0], return_counts=True)
        split_idx = np.cumsum(samples_per_run)[:-1]
        features = np.split(raw_features, split_idx, axis=0)
        cond_per_run = np.split(operation_conditions, split_idx, axis=0)

        return features, cond_per_run

    def _scale_features(
        self, features: List[np.ndarray], operation_conditions: List[np.ndarray]
    ) -> List[np.ndarray]:
        scaler_path = self._get_scaler_path()
        if not os.path.exists(scaler_path):
            raise RuntimeError(
                f"Scaler for FD{self.fd:03d} with features "
                f"{self.feature_select} does not exist. "
                f"Did you call prepare_data yet?"
            )
        scaler = scaling.load_scaler(scaler_path)
        if self.operation_condition_aware_scaling:
            operation_conditions = [
                c[:, self._CONDITION_COLUMN] for c in operation_conditions
            ]
            features = scaling.scale_features(features, scaler, operation_conditions)
        else:
            features = scaling.scale_features(features, scaler)

        return features

    @staticmethod
    def _split_time_steps_from_features(
        features: List[np.ndarray],
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        time_steps = []
        for i, seq in enumerate(features):
            time_steps.append(seq[:, 1])
            seq = seq[:, 2:]
            features[i] = seq

        return features, time_steps

    def _generate_targets(self, time_steps: List[np.ndarray]) -> List[np.ndarray]:
        """Generate RUL targets from time steps."""
        max_rul = self.max_rul or np.inf  # no capping if max_rul is None
        targets = [np.minimum(max_rul, steps)[::-1].copy() for steps in time_steps]

        return targets

    def _load_targets(self, features: List[np.ndarray]) -> List[np.ndarray]:
        """Load target file."""
        file_name = f"RUL_FD{self.fd:03d}.txt"
        file_path = os.path.join(self._CMAPSS_ROOT, file_name)
        raw_targets = np.loadtxt(file_path)

        targets = np.split(raw_targets, len(raw_targets))
        targets = [np.arange(len(f), 0, -1) + t - 1 for f, t in zip(features, targets)]

        max_rul = self.max_rul or np.inf
        targets = [np.minimum(max_rul, t) for t in targets]

        return targets

    def _pad_data(self, features, targets):
        padded_features = []
        padded_targets = []
        for i, (feat, target) in enumerate(zip(features, targets)):
            if feat.shape[0] < self.window_size:
                warnings.warn(
                    f"Run {i} of CMAPSS FD{self.fd:03d} is shorter than window "
                    f"size {self.window_size} and will be zero-padded."
                )
                missing = self.window_size - feat.shape[0]
                feat_pad = (missing, feat.shape[1])
                feat = np.concatenate([np.zeros(feat_pad), feat])
                target = np.concatenate([np.zeros(missing), target])
            padded_features.append(feat)
            padded_targets.append(target)

        return padded_features, padded_targets

    def _window_data(
        self, features: List[np.ndarray], targets: List[np.ndarray]
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Window features with specified window size."""
        windowed_features = []
        windowed_targets = []
        for seq, target in zip(features, targets):
            windows = utils.extract_windows(seq, self.window_size)
            target = target[self.window_size - 1 :]
            windowed_features.append(windows)
            windowed_targets.append(target)

        return windowed_features, windowed_targets

    def _crop_data(
        self, features: List[np.ndarray], targets: List[np.ndarray]
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """Crop length of data to specified window size."""
        cropped_features = []
        cropped_targets = []
        for seq, target in zip(features, targets):
            cropped_features.append(seq[None, -self.window_size :])
            cropped_targets.append(target[-1, None])

        return cropped_features, cropped_targets


def _download_cmapss(data_root: str) -> None:
    with tempfile.TemporaryDirectory() as tmp_path:
        print("Download CMAPSS dataset")
        download_path = os.path.join(tmp_path, "CMAPSSData.zip")
        utils.download_file(CMAPSS_URL, download_path)
        print("Extract CMAPSS dataset")
        with zipfile.ZipFile(download_path, mode="r") as f:
            f.extractall(data_root)



"""Basic data modules for experiments involving only a single subset of any RUL
dataset. """

from typing import Dict, List, Optional, Tuple, Any, Callable

import numpy as np
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader, IterableDataset, TensorDataset, get_worker_info

from rul_datasets import utils
from rul_datasets.reader import AbstractReader


class RulDataModule(pl.LightningDataModule):
    """
    A [data module][pytorch_lightning.core.LightningDataModule] to provide windowed
    time series features with RUL targets. It exposes the splits of the underlying
    dataset for easy usage with PyTorch and PyTorch Lightning.

    The data module implements the `hparams` property used by PyTorch Lightning to
    save hyperparameters to checkpoints. It retrieves the hyperparameters of its
    underlying reader and adds the batch size to them.

    If you want to extract features from the windows, you can pass the
    `feature_extractor` and `window_size` arguments to the constructor. The
    `feature_extractor` is a callable that takes a windowed time series as a numpy
    array with the shape `[num_windows, window_size, num_features]` and returns
    another numpy array. Depending on `window_size`, the expected output shapes for
    the `feature_extractor` are:

    * `window_size is None`: `[num_new_windows, new_window_size, features]`
    * `window_size is not None`: `[num_windows, features]`

    If `window_size` is set, the extracted features are re-windowed.

    Examples:
        Default

        >>> import rul_datasets
        >>> cmapss = rul_datasets.reader.CmapssReader(fd=1)
        >>> dm = rul_datasets.RulDataModule(cmapss, batch_size=32)

        With Feature Extractor

        >>> import rul_datasets
        >>> import numpy as np
        >>> cmapss = rul_datasets.reader.CmapssReader(fd=1)
        >>> dm = rul_datasets.RulDataModule(
        ...     cmapss,
        ...     batch_size=32,
        ...     feature_extractor=lambda x: np.mean(x, axis=1),
        ...     window_size=10
        ... )
    """

    _data: Dict[str, Tuple[torch.Tensor, torch.Tensor]]

    def __init__(
        self,
        reader: AbstractReader,
        batch_size: int,
        feature_extractor: Optional[Callable] = None,
        window_size: Optional[int] = None,
    ):
        """
        Create a new RUL data module from a reader.

        This data module exposes a training, validation and test data loader for the
        underlying dataset. First, `prepare_data` is called to download and
        pre-process the dataset. Afterwards, `setup_data` is called to load all
        splits into memory.

        If a `feature_extractor` is supplied, the data module extracts new features
        from each window of the time series. If `window_size` is `None`,
        it is assumed that the extracted features form a new windows themselves. If
        `window_size` is an int, it is assumed that the extracted features are a
        single feature vectors and should be re-windowed. The expected output shapes
        for the `feature_extractor` are:

        * `window_size is None`: `[num_new_windows, new_window_size, features]`
        * `window_size is not None`: `[num_windows, features]`

        The expected input shape for the `feature_extractor` is always
        `[num_windows, window_size, features]`.

        Args:
            reader: The dataset reader for the desired dataset, e.g. CmapssLoader.
            batch_size: The size of the batches build by the data loaders.
            feature_extractor: A feature extractor that extracts feature vectors from
                               windows.
            window_size: The new window size to apply after the feature extractor.
        """
        super().__init__()

        self._reader: AbstractReader = reader
        self.batch_size = batch_size
        self.feature_extractor = feature_extractor
        self.window_size = window_size

        if (self.feature_extractor is None) and (self.window_size is not None):
            raise ValueError(
                "A feature extractor has to be supplied "
                "to set a window size for re-windowing."
            )

        hparams = {
            "reader": self.reader.hparams,
            "batch_size": self.batch_size,
            "feature_extractor": (
                str(self.feature_extractor) if self.feature_extractor else None
            ),
            "window_size": self.window_size,
        }
        self.save_hyperparameters(hparams)

    @property
    def data(self) -> Dict[str, Tuple[torch.Tensor, torch.Tensor]]:
        """
        A dictionary of the training, validation and test splits.

        Each split is a tuple of feature and target tensors.
        The keys are `dev` (training split), `val` (validation split) and `test`
        (test split).
        """
        return self._data

    @property
    def reader(self) -> AbstractReader:
        """The underlying dataset reader."""
        return self._reader

    @property
    def fds(self):
        """Index list of the available subsets of the underlying dataset, i.e.
        `[1, 2, 3, 4]` for `CMAPSS`."""
        return self._reader.fds

    def check_compatibility(self, other: "RulDataModule") -> None:
        """
        Check if another RulDataModule is compatible to be used together with this one.

        RulDataModules can be used together in higher-order data modules,
        e.g. AdaptionDataModule. This function checks if `other` is compatible to
        this data module to do so. It checks the underlying dataset readers, matching
        batch size, feature extractor and window size. If anything is incompatible,
        this function will raise a ValueError.

        Args:
            other: The RulDataModule to check compatibility with.
        """
        try:
            self.reader.check_compatibility(other.reader)
        except ValueError:
            raise ValueError("RulDataModules incompatible on reader level.")

        if not self.batch_size == other.batch_size:
            raise ValueError(
                f"The batch size of both data modules has to be the same, "
                f"{self.batch_size} vs. {other.batch_size}."
            )

        if not self.window_size == other.window_size:
            raise ValueError(
                f"The window size of both data modules has to be the same, "
                f"{self.window_size} vs. {other.window_size}."
            )

    def is_mutually_exclusive(self, other: "RulDataModule") -> bool:
        """
        Check if the other data module is mutually exclusive to this one. See
        [AbstractReader.is_mutually_exclusive]
        [rul_datasets.reader.abstract.AbstractReader.is_mutually_exclusive].

        Args:
            other: Data module to check exclusivity against.
        Returns:
            Whether both data modules are mutually exclusive.
        """
        return self.reader.is_mutually_exclusive(other.reader)

    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        """
        Download and pre-process the underlying data.

        This calls the `prepare_data` function of the underlying reader. All
        previously completed preparation steps are skipped. It is called
        automatically by `pytorch_lightning` and executed on the first GPU in
        distributed mode.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        """
        self.reader.prepare_data()

    def setup(self, stage: Optional[str] = None) -> None:
        """
        Load all splits as tensors into memory and optionally apply feature extractor.

        The splits are placed inside the [data][rul_datasets.core.RulDataModule.data]
        property. If a split is empty, a tuple of empty tensors with the correct
        number of dimensions is created as a placeholder. This ensures compatibility
        with higher-order data modules.

        If the data module was constructed with a `feature_extractor` argument,
        the feature windows are passed to the feature extractor. The resulting,
        new features may be re-windowed.

        Args:
            stage: Ignored. Only for adhering to parent class interface.
        """
        self._data = {
            "dev": self._setup_split("dev"),
            "val": self._setup_split("val"),
            "test": self._setup_split("test"),
        }

    def load_split(
        self, split: str, alias: Optional[str] = None
    ) -> Tuple[List[torch.Tensor], List[torch.Tensor]]:
        """
        Load a split from the underlying reader and apply the feature extractor.

        By setting alias, it is possible to load a split aliased as another split,
        e.g. load the test split and treat it as the dev split. The data of the split is
        loaded but all pre-processing steps of alias are carried out.

        Args:
            split: The desired split to load.
            alias: The split as which the loaded data should be treated.
        Returns:
            The feature and target tensors of the split's runs.
        """
        features, targets = self.reader.load_split(split, alias)
        features, targets = self._apply_feature_extractor_per_run(features, targets)
        tensor_features, tensor_targets = utils.to_tensor(features, targets)

        return tensor_features, tensor_targets

    def _setup_split(
        self, split: str, alias: Optional[str] = None
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        features, targets = self.load_split(split, alias)
        if features:
            cat_features, cat_targets = torch.cat(features), torch.cat(targets)
        else:
            cat_features, cat_targets = torch.empty(0, 0, 0), torch.empty(0)

        return cat_features, cat_targets

    def _apply_feature_extractor_per_run(
        self, features: List[np.ndarray], targets: List[np.ndarray]
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        extracted = [self._extract_and_window(f, t) for f, t in zip(features, targets)]
        features, targets = zip(*extracted) if extracted else ((), ())

        return list(features), list(targets)

    def _extract_and_window(
        self, features: np.ndarray, targets: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
        if self.feature_extractor is not None:
            features, targets = self.feature_extractor(features, targets)
        if self.window_size is not None:
            cutoff = self.window_size - 1
            features = utils.extract_windows(features, self.window_size)
            targets = targets[cutoff:]

        return features, targets

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """
        Create a [data loader][torch.utils.data.DataLoader] for the training split.

        The data loader is configured to shuffle the data. The `pin_memory` option is
        activated to achieve maximum transfer speed to the GPU. The data loader is also
        configured to drop the last batch of the data if it would only contain one
        sample.

        The whole split is held in memory. Therefore, the `num_workers` are set to
        zero which uses the main process for creating batches.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The training data loader
        """
        dataset = self.to_dataset("dev")
        drop_last = len(dataset) % self.batch_size == 1
        loader = DataLoader(
            dataset,
            batch_size=self.batch_size,
            shuffle=True,
            drop_last=drop_last,
            pin_memory=True,
        )

        return loader

    def val_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """
        Create a [data loader][torch.utils.data.DataLoader] for the validation split.

        The data loader is configured to leave the data unshuffled. The `pin_memory`
        option is activated to achieve maximum transfer speed to the GPU.

        The whole split is held in memory. Therefore, the `num_workers` are set to
        zero which uses the main process for creating batches.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The validation data loader
        """
        return DataLoader(
            self.to_dataset("val"),
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
        )

    def test_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """
        Create a [data loader][torch.utils.data.DataLoader] for the test split.

        The data loader is configured to leave the data unshuffled. The `pin_memory`
        option is activated to achieve maximum transfer speed to the GPU.

        The whole split is held in memory. Therefore, the `num_workers` are set to
        zero which uses the main process for creating batches.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The test data loader
        """
        return DataLoader(
            self.to_dataset("test"),
            batch_size=self.batch_size,
            shuffle=False,
            pin_memory=True,
        )

    def to_dataset(self, split: str, alias: Optional[str] = None) -> TensorDataset:
        """
        Create a dataset of a split.

        This convenience function creates a plain [tensor dataset]
        [torch.utils.data.TensorDataset] to use outside the `rul_datasets` library.

        The data placed inside the dataset will be from the specified `split`. If
        `alias` is set, the loaded data will be treated as if from the `alias` split.
        For example, one could load the test data and treat them as if it was the
        training data. This may be useful for inductive domain adaption.

        Args:
            split: The split to place inside the dataset.
            alias: The split the loaded data should be treated as.
        Returns:
            A dataset containing the requested split.
        """
        if (alias is None) or (split == alias):
            features, targets = self._data[split]
        else:
            features, targets = self._setup_split(split, alias)
        split_dataset = TensorDataset(features, targets)

        return split_dataset


class PairedRulDataset(IterableDataset):
    """A dataset of sample pairs drawn from the same time series."""

    def __init__(
        self,
        readers: List[AbstractReader],
        split: str,
        num_samples: int,
        min_distance: int,
        deterministic: bool = False,
        mode: str = "linear",
    ):
        super().__init__()

        self.readers = readers
        self.split = split
        self.min_distance = min_distance
        self.num_samples = num_samples
        self.deterministic = deterministic
        self.mode = mode

        for reader in self.readers:
            reader.check_compatibility(self.readers[0])

        self._run_domain_idx: np.ndarray
        self._features: List[np.ndarray]
        self._labels: List[np.ndarray]
        self._prepare_datasets()

        self._max_rul = self._get_max_rul()
        self._curr_iter = 0
        self._rng = self._reset_rng()
        if mode == "linear":
            self._get_pair_func = self._get_pair_idx
        elif mode == "piecewise":
            self._get_pair_func = self._get_pair_idx_piecewise
        elif mode == "labeled":
            self._get_pair_func = self._get_labeled_pair_idx

    def _get_max_rul(self):
        max_ruls = [reader.max_rul for reader in self.readers]
        if any(m is None for m in max_ruls):
            raise ValueError(
                "PairedRulDataset needs a set max_rul for all readers "
                "but at least one of them has is None."
            )
        max_rul = max(max_ruls)

        return max_rul

    def _prepare_datasets(self):
        run_domain_idx = []
        features = []
        labels = []
        for domain_idx, reader in enumerate(self.readers):
            run_features, run_labels = reader.load_split(self.split)
            for feat, lab in zip(run_features, run_labels):
                if len(feat) > self.min_distance:
                    run_domain_idx.append(domain_idx)
                    features.append(feat)
                    labels.append(lab)

        self._run_domain_idx = np.array(run_domain_idx)
        self._features = features
        self._labels = labels

    def _reset_rng(self, seed=42) -> np.random.Generator:
        return np.random.default_rng(seed=seed)

    def __len__(self) -> int:
        return self.num_samples

    def __iter__(self):
        self._curr_iter = 0
        worker_info = get_worker_info()
        if self.deterministic and worker_info is not None:
            raise RuntimeError(
                "PairedDataset cannot run deterministic in multiprocessing"
            )
        elif self.deterministic:
            self._rng = self._reset_rng()
        elif worker_info is not None:
            self._rng = self._reset_rng(worker_info.seed)

        return self

    def __next__(self) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        if self._curr_iter < self.num_samples:
            run_idx, anchor_idx, query_idx, dist, domain_label = self._get_pair_func()
            self._curr_iter += 1
            run = self._features[run_idx]
            return self._build_pair(run, anchor_idx, query_idx, dist, domain_label)
        else:
            raise StopIteration

    def _get_pair_idx(self) -> Tuple[int, int, int, int, int]:
        chosen_run_idx = self._rng.integers(0, len(self._features))
        domain_label = self._run_domain_idx[chosen_run_idx]
        chosen_run = self._features[chosen_run_idx]

        run_length = chosen_run.shape[0]
        anchor_idx = self._rng.integers(
            low=0,
            high=run_length - self.min_distance,
        )
        end_idx = min(run_length, anchor_idx + self._max_rul)
        query_idx = self._rng.integers(
            low=anchor_idx + self.min_distance,
            high=end_idx,
        )
        distance = query_idx - anchor_idx

        return chosen_run_idx, anchor_idx, query_idx, distance, domain_label

    def _get_pair_idx_piecewise(self) -> Tuple[int, int, int, int, int]:
        chosen_run_idx = self._rng.integers(0, len(self._features))
        domain_label = self._run_domain_idx[chosen_run_idx]
        chosen_run = self._features[chosen_run_idx]

        run_length = chosen_run.shape[0]
        middle_idx = run_length // 2
        anchor_idx = self._rng.integers(
            low=0,
            high=run_length - self.min_distance,
        )
        end_idx = (
            middle_idx if anchor_idx < (middle_idx - self.min_distance) else run_length
        )
        query_idx = self._rng.integers(
            low=anchor_idx + self.min_distance,
            high=end_idx,
        )
        distance = query_idx - anchor_idx if anchor_idx > middle_idx else 0

        return chosen_run_idx, anchor_idx, query_idx, distance, domain_label

    def _get_labeled_pair_idx(self) -> Tuple[int, int, int, int, int]:
        chosen_run_idx = self._rng.integers(0, len(self._features))
        domain_label = self._run_domain_idx[chosen_run_idx]
        chosen_run = self._features[chosen_run_idx]
        chosen_labels = self._labels[chosen_run_idx]

        run_length = chosen_run.shape[0]
        anchor_idx = self._rng.integers(
            low=0,
            high=run_length - self.min_distance,
        )
        query_idx = self._rng.integers(
            low=anchor_idx + self.min_distance,
            high=run_length,
        )
        # RUL label difference is negative time step difference
        distance = int(chosen_labels[anchor_idx] - chosen_labels[query_idx])

        return chosen_run_idx, anchor_idx, query_idx, distance, domain_label

    def _build_pair(
        self,
        run: np.ndarray,
        anchor_idx: int,
        query_idx: int,
        distance: int,
        domain_label: int,
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        anchors = utils.feature_to_tensor(run[anchor_idx], torch.float)
        queries = utils.feature_to_tensor(run[query_idx], torch.float)
        domain_tensor = torch.tensor(domain_label, dtype=torch.float)
        distances = torch.tensor(distance, dtype=torch.float) / self._max_rul
        distances = torch.clamp_max(distances, max=1)  # max distance is max_rul

        return anchors, queries, distances, domain_tensor



"""A module with functions for truncating RUL data."""

from typing import List, Tuple, Iterable, Union, Optional

import numpy as np


def truncate_runs(
    features: List[np.ndarray],
    targets: List[np.ndarray],
    percent_broken: Optional[float] = None,
    included_runs: Optional[Union[float, Iterable[int]]] = None,
    degraded_only: bool = False,
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    Truncate RUL data according to `percent_broken` and `included_runs`.

    RUL data has two dimensions in which it can be truncated: the number of runs and
    the length of the runs. Truncating the number of runs limits the inter-run
    variety of the data. Truncating the length of the run limits the amount of
    available data near failure.

    For more information about truncation, see the [reader][rul_datasets.reader]
    module page.

    Examples:
        Truncating via `percent_broken`
        >>> import numpy as np
        >>> from rul_datasets.reader.truncating import truncate_runs
        >>> features = [np.random.randn(i*100, 5) for i in range(1, 6)]
        >>> targets = [np.arange(i*100)[::-1] for i in range(1, 6)]
        >>> (features[0].shape, targets[0].shape)
        ((100, 5), (100,))
        >>> features, targets = truncate_runs(features, targets, percent_broken=0.8)
        >>> (features[0].shape, targets[0].shape)  # runs are shorter
        ((80, 5), (80,))
        >>> np.min(targets[0])  # runs contain no failures
        20
    """
    # Truncate the number of runs
    if included_runs is not None:
        features, targets = _truncate_included(features, targets, included_runs)
    # Truncate the number of samples per run, starting at failure
    if percent_broken is not None and percent_broken < 1:
        features, targets = _truncate_broken(
            features, targets, percent_broken, degraded_only
        )

    return features, targets


def _truncate_included(
    features: List[np.ndarray],
    targets: List[np.ndarray],
    included_runs: Union[float, Iterable[int]],
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    if isinstance(included_runs, float):
        features, targets = _truncate_included_by_percentage(
            features, targets, included_runs
        )
    elif isinstance(included_runs, Iterable):
        features, targets = _truncate_included_by_index(
            features, targets, included_runs
        )

    return features, targets


def _truncate_broken(
    features: List[np.ndarray],
    targets: List[np.ndarray],
    percent_broken: float,
    degraded_only: bool,
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    features = features.copy()  # avoid mutating original list
    targets = targets.copy()  # avoid mutating original list
    for i, (run, target) in enumerate(zip(features, targets)):
        if degraded_only:
            num_healthy = np.sum(target == np.max(target))
            num_degraded = len(run) - num_healthy
            num_cycles = num_healthy + int(percent_broken * num_degraded)
        else:
            num_cycles = int(percent_broken * len(run))
        features[i] = run[:num_cycles]
        targets[i] = target[:num_cycles]

    return features, targets


def _truncate_included_by_index(
    features: List[np.ndarray], targets: List[np.ndarray], included_idx: Iterable[int]
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    features = [features[i] for i in included_idx]
    targets = [targets[i] for i in included_idx]

    return features, targets


def _truncate_included_by_percentage(
    features: List[np.ndarray], targets: List[np.ndarray], percent_included: float
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    num_runs = int(percent_included * len(features))
    features = features[:num_runs]
    targets = targets[:num_runs]

    return features, targets



"""A module for dataset readers. Currently supported datasets are:

* **C-MAPSS** Turbofan Degradation Dataset
* **FEMTO** (PRONOSTIA) Bearing Dataset
* **XJTU-SY** Bearing Dataset

Readers are the foundation of the RUL Datasets library. They provide access to the
data on disk and convert them into a common format so that other parts of the library
can interact with it. The common format is as follows:

1. Each dataset consists of multiple sub-datasets. The indices of these sub-datasets
are called `FD`, following CMAPSS convention.

1. Each sub-dataset contains a development (`dev`), a validation (`val`) and test split
(`test`).

1. Each split contains one or multiple time series of features and RUL targets that
represent run-to-failure experiments.

1. At each time step of a time series we have a window of features and a target RUL
value. The target is the RUL value of the last time step of the window.

A reader class, e.g. the [CmapssReader][rul_datasets.reader.cmapss.CmapssReader]
represents a dataset and can manipulate it to your liking. A reader object has access
to one sub-dataset of the dataset:

```pycon
>>> reader = CmapssReader(fd=1)
```

The reader object can load the features and targets of each split into memory:

```pycon
>>> dev_features, dev_targets = reader.load_split("dev")
>>> val_features, val_targets = reader.load_split("val")
>>> test_features, test_targets = reader.load_split("test")
```

The features are a list of [numpy arrays][numpy.ndarray] where each array has a shape of
`[num_windows, window_size, num_channels]`:

```pycon
>>> type(dev_features)
<class 'list'>
>>> dev_features[0].shape
(163, 30, 14)
```

The targets are a list of [numpy arrays][numpy.ndarray], too, where each array has a
shape of `[num_windows]`:

```pycon
>>> type(dev_targets)
<class 'list'>
>>> dev_targets[0].shape
(163,)
```

Each reader defines a default window size for its data. This can be overridden by the
`window_size` argument:

```pycon
>>> fd1 = CmapssReader(fd=1, window_size=15)
>>> features, _ = fd1.load_split("dev")
>>> features[0].shape
(163, 15, 14)
```

Some datasets, i.e. CMAPSS, use a piece-wise linear RUL function, where a maximum RUL
value is defined. The maximum RUL value for a reader can be set via the `max_rul`
argument:

```pycon
>>> fd1 = CmapssReader(fd=1, max_rul=100)
>>> targets = fd1.load_split("dev")
>>> max(np.max(t) for t in targets)
100.0
```

If you want to use a sub-dataset as unlabeled data, e.g. for unsupervised domain
adaption, it should not contain features from the point of failure. If the data
contains these features, there would be no reason for it to be unlabeled. The
`percent_broken` argument controls how much data near failure is available. A
`percent_broken` of `0.8` for example means that only the first 80% of each time
series are available:

```pycon
>>> fd1 = CmapssReader(fd=1, percent_broken=0.8)
>>> features, targets = fd1.load_split("dev")
>>> features[0].shape
(130, 30, 14])
>>> np.min(targets[0])
34.0
```

If you have set a `max_rul` you may only want to truncate data that is considered
degraded, i.e. with a RUL value smaller than `max_rul`. You can use the
`truncate_degraded_only` option to do that. This way, the data where the RUL value is
smaller or equal to `(1 - percent_broken) * max_rul` is cut off.

```pycon
>>> fd1 = CmapssReader(fd=1, percent_broken=0.8, truncate_degraded_only=True)
>>> features, targets = fd1.load_split("dev")
>>> features[0].shape
(138, 30, 14])
>>> np.min(targets[0])
26.0
```

You may want to apply the same `percent_broken` from your training data to your
validation data. This is sensible if you do not expect that your algorithm has access
to labeled validation data in real-life. You can achieve this, by setting
`truncate_val` to `True`:

```pycon
>>> fd1 = CmapssReader(fd=1, percent_broken=0.8, truncate_val=True)
>>> features, targets = fd1.load_split("val")
>>> np.min(targets[0])
44.0
```

Data-driven RUL estimation algorithms are often sensitive to the overall amount of
training data. The more data is available, the more of its variance is covered. If
you want to investigate how an algorithm performs in a low-data setting, you can use
`percent_fail_runs`. This argument controls how many runs are used for training. A
`percent_fail_runs` of `0.8` means that 80% of the available training runs are used.
If you need more controll over which runs are used, you can pass a list of indices to
use only these runs. This is useful for conducting semi-supervised learning where you
consider one part of a sub-dataset labeled and the other part unlabeled:

```pycon
>>> fd1 = CmapssReader(fd=1, percent_fail_runs=0.8)
>>> features, targets = fd1.load_split("dev")
>>> len(features)
64
>>> fd1 = CmapssReader(fd=1, percent_fail_runs=[0, 5, 40])
>>> features, targets = fd1.load_split("dev")
>>> len(features)
3
```

If you have constructed a reader with a certain `percent_fail_runs`, you can get a
reader containing all other runs by using the `get_complement` function:

```pycon
>>> fd1 = CmapssReader(fd=1, percent_fail_runs=0.8)
>>> fd1_complement = fd1.get_complement()
>>> features, targets = fd1_complement.load_split("dev")
>>> len(features)
16
```

The effects of `percent_broken` and `percent_fail_runs` are summarized under the term
**truncation** as they effectively truncate the dataset in two dimensions.

The readers for the FEMTO and XJTU-SY datasets have two additional constructor
arguments. The `first_time_to_predict` lets you set an individual maximum RUL value
per run in the dataset. As both are bearing datasets, the first-time-to-predict is
defined as the time step where the degradation of the bearing is first noticeable.
The RUL value before this time step is assumed to be constant. Setting `norm_rul`
scales the RUL between [0, 1] per run, as it is best practice when using
first-time-to-predict.

```pycon
>>> fttp = [10, 20, 30, 40, 50]
>>> fd1 = rul_datasets.reader.XjtuSyReader(
...     fd=1, first_time_to_predict=fttp, norm_rul=True
... )
>>> fd1.prepare_data()
>>> features, labels = fd1.load_split("dev")
>>> labels[0][:15]
array([1.        , 1.        , 1.        , 1.        , 1.        ,
       1.        , 1.        , 1.        , 1.        , 1.        ,
       1.        , 0.99115044, 0.98230088, 0.97345133, 0.96460177])
```

Readers can be used as is if you just want access to the dataset. If you plan to use
them with PyTorch or PyTorch Lightning, it is recommended to combine them with a
[RulDataModule][rul_datasets.core.RulDataModule]:

```pycon
>>> fd1 = CmapssReader(fd=1)
>>> dm = RulDataModule(fd1, batch_size=32)
```

For more information, see [core][rul_datasets.core] module page or the
[Libraries](/rul-datasets/use_cases/libraries) page.

 """

from .abstract import AbstractReader
from .data_root import _DATA_ROOT
from .cmapss import CmapssReader
from .femto import FemtoReader, FemtoPreparator
from .xjtu_sy import XjtuSyReader, XjtuSyPreparator
from .dummy import DummyReader



"""The FEMTO (PRONOSTIA) Bearing dataset is a collection of run-to-failure
experiments on bearings. Three different operation conditions were used, resulting in
three sub-datasets. Sub-dataset 1 and 2 contain two training runs and five test runs,
while sub-dataset 3 contains only one test run. It was part of the 2012 IEEE
Prognostics Challenge. """

import os
import re
import tempfile
import warnings
import zipfile
from typing import List, Tuple, Union, Dict, Optional

import numpy as np
import sklearn.preprocessing as scalers  # type: ignore

from rul_datasets import utils
from rul_datasets.reader import scaling, saving
from rul_datasets.reader.data_root import get_data_root
from rul_datasets.reader.abstract import AbstractReader

FEMTO_URL = "https://kr0k0tsch.de/rul-datasets/FEMTOBearingDataSet.zip"


class FemtoReader(AbstractReader):
    """
    This reader represents the FEMTO (PRONOSTIA) Bearing dataset. Each of its three
    sub-datasets contains a training and a test split. By default, the reader
    constructs a validation split for sub-datasets 1 and 2 each by taking the first
    run of the test split. For sub-dataset 3, the second training run is used for
    validation because only one test run is available. The remaining training data is
    denoted as the development split. This run to split assignment can be overridden
    by setting `run_split_dist`.

    The features contain windows with three channels. Only the two acceleration
    channels are used because the test runs are missing the temperature channel.
    These features are standardized to zero mean and one standard deviation. The
    scaler is fitted on the development data.

    Examples:
        Default splits:
        >>> import rul_datasets
        >>> fd1 = rul_datasets.reader.FemtoReader(fd=1)
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> features[0].shape
        (2803, 2560, 2)

        Custom splits:
        >>> import rul_datasets
        >>> splits = {"dev": [5], "val": [4], "test": [3]}
        >>> fd1 = rul_datasets.reader.FemtoReader(fd=1, run_split_dist=splits)
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> features[0].shape
        (2463, 2560, 2)

        Set first-time-to-predict:
        >>> import rul_datasets
        >>> fttp = [10, 20, 30, 40, 50]
        >>> fd1 = rul_datasets.reader.FemtoReader(fd=1, first_time_to_predict=fttp)
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> labels[0][:15]
        array([2793., 2793., 2793., 2793., 2793., 2793., 2793., 2793., 2793.,
               2793., 2793., 2792., 2791., 2790., 2789.])
    """

    _FEMTO_ROOT: str = os.path.join(get_data_root(), "FEMTOBearingDataSet")
    _NUM_TRAIN_RUNS: Dict[int, int] = {1: 2, 2: 2, 3: 2}

    def __init__(
        self,
        fd: int,
        window_size: Optional[int] = None,
        max_rul: Optional[int] = None,
        percent_broken: Optional[float] = None,
        percent_fail_runs: Optional[Union[float, List[int]]] = None,
        truncate_val: bool = False,
        run_split_dist: Optional[Dict[str, List[int]]] = None,
        first_time_to_predict: Optional[List[int]] = None,
        norm_rul: bool = False,
        truncate_degraded_only: bool = False,
    ) -> None:
        """
        Create a new FEMTO reader for one of the sub-datasets. By default, the RUL
        values are not capped. The default window size is 2560.

         Use `first_time_to_predict` to set an individual RUL inflection point for
        each run. It should be a list with an integer index for each run. The index
        is the time step after which RUL declines. Before the time step it stays
        constant. The `norm_rul` argument can then be used to scale the RUL of each
        run between zero and one.

        For more information about using readers refer to the [reader]
        [rul_datasets.reader] module page.

        Args:
            fd: Index of the selected sub-dataset
            window_size: Size of the sliding window. Defaults to 2560.
            max_rul: Maximum RUL value of targets.
            percent_broken: The maximum relative degradation per time series.
            percent_fail_runs: The percentage or index list of available time series.
            truncate_val: Truncate the validation data with `percent_broken`, too.
            run_split_dist: Dictionary that assigns each run idx to each split.
            first_time_to_predict: The time step for each time series before which RUL
                                   is constant.
            norm_rul: Normalize RUL between zero and one.
            truncate_degraded_only: Only truncate the degraded part of the data
                                    (< max RUL).
        """
        super().__init__(
            fd,
            window_size,
            max_rul,
            percent_broken,
            percent_fail_runs,
            truncate_val,
            truncate_degraded_only,
        )

        if (first_time_to_predict is not None) and (max_rul is not None):
            raise ValueError(
                "FemtoReader cannot use 'first_time_to_predict' "
                "and 'max_rul' in conjunction."
            )

        self.first_time_to_predict = first_time_to_predict
        self.norm_rul = norm_rul

        self._preparator = FemtoPreparator(self.fd, self._FEMTO_ROOT, run_split_dist)

    @property
    def dataset_name(self) -> str:
        return "femto"

    @property
    def fds(self) -> List[int]:
        """Indices of available sub-datasets."""
        return list(self._NUM_TRAIN_RUNS)

    def prepare_data(self) -> None:
        """
        Prepare the FEMTO dataset. This function needs to be called before using the
        dataset and each custom split for the first time.

        The dataset is downloaded from a custom mirror and extracted into the data
        root directory. The whole dataset is converted from CSV files to NPY files to
        speed up loading it from disk. Afterwards, a scaler is fit on the development
        features. Previously completed steps are skipped.
        """
        if not os.path.exists(self._FEMTO_ROOT):
            _download_femto(get_data_root())
        self._preparator.prepare_split("dev")
        self._preparator.prepare_split("val")
        self._preparator.prepare_split("test")

    def load_complete_split(
        self, split: str, alias: str
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        features, targets = self._preparator.load_runs(split)
        features = [f[:, -self.window_size :, :] for f in features]  # crop to window
        features = scaling.scale_features(features, self._preparator.load_scaler())
        if self.max_rul is not None:
            targets = [np.minimum(t, self.max_rul) for t in targets]
        elif self.first_time_to_predict is not None:
            targets = self._clip_first_time_to_predict(targets, split)

        if self.norm_rul:
            targets = [t / np.max(t) for t in targets]

        return features, targets

    def _clip_first_time_to_predict(self, targets, split):
        fttp = [
            self.first_time_to_predict[i - 1]
            for i in self._preparator.run_split_dist[split]
        ]
        targets = [np.minimum(t, len(t) - fttp) for t, fttp in zip(targets, fttp)]

        return targets

    def default_window_size(self, fd: int) -> int:
        return FemtoPreparator.DEFAULT_WINDOW_SIZE


class FemtoPreparator:
    DEFAULT_WINDOW_SIZE = 2560
    _SPLIT_FOLDERS = {
        "dev": "Learning_set",
        "val": "Full_Test_Set",
        "test": "Full_Test_Set",
    }
    _DEFAULT_RUN_SPLIT_DIST = {
        1: {"dev": [1, 2], "val": [3], "test": [4, 5, 6, 7]},
        2: {"dev": [1, 2], "val": [3], "test": [4, 5, 6, 7]},
        3: {"dev": [1], "val": [2], "test": [3]},
    }

    def __init__(
        self,
        fd: int,
        data_root: str,
        run_split_dist: Optional[Dict[str, List[int]]] = None,
    ) -> None:
        self.fd = fd
        self._data_root = data_root
        self.run_split_dist = run_split_dist or self._DEFAULT_RUN_SPLIT_DIST[self.fd]

    def prepare_split(self, split: str) -> None:
        if not self._split_already_prepared(split):
            warnings.warn(f"First time use. Pre-process {split} split of FD{self.fd}.")
            runs = self._load_raw_runs(split)
            self._save_efficient(split, runs)
        if split == "dev" and not os.path.exists(self._get_scaler_path()):
            features, _ = self.load_runs(split)
            scaler = scaling.fit_scaler(features)
            scaling.save_scaler(scaler, self._get_scaler_path())

    def _split_already_prepared(self, split: str) -> bool:
        run_idx_in_split = self._DEFAULT_RUN_SPLIT_DIST[self.fd][split][0]
        run_file_path = self._get_run_file_path(split, run_idx_in_split)
        already_prepared = saving.exists(run_file_path)

        return already_prepared

    def load_runs(self, split: str) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        self._validate_split(split)
        run_idx = self.run_split_dist[split]
        run_paths = [self._get_run_file_path(split, idx) for idx in run_idx]
        features, targets = saving.load_multiple(run_paths)

        return features, targets

    def load_scaler(self) -> scalers.StandardScaler:
        return scaling.load_scaler(self._get_scaler_path())

    def _load_raw_runs(self, split: str) -> Dict[int, Tuple[np.ndarray, np.ndarray]]:
        file_paths = self._get_csv_file_paths(split)
        features = saving.load_raw(file_paths, self.DEFAULT_WINDOW_SIZE, columns=[4, 5])
        targets = utils.get_targets_from_file_paths(
            file_paths, self._timestep_from_file_path
        )
        runs = {idx: (features[idx], targets[idx]) for idx in features}

        return runs

    def _get_csv_file_paths(self, split: str) -> Dict[int, List[str]]:
        split_path = self._get_split_folder(split)
        run_folders = self._get_run_folders(split_path)
        file_paths = {}
        for run_idx, run_folder in run_folders.items():
            run_path = os.path.join(split_path, run_folder)
            feature_files = utils.get_files_in_path(
                run_path, lambda f: f.startswith("acc")
            )
            file_paths[run_idx] = feature_files

        return file_paths

    @staticmethod
    def _timestep_from_file_path(file_path: str) -> int:
        file_name = os.path.basename(file_path)
        time_step = int(file_name[4:9])

        return time_step

    def _save_efficient(
        self, split: str, runs: Dict[int, Tuple[np.ndarray, np.ndarray]]
    ) -> None:
        for run_idx, (features, targets) in runs.items():
            saving.save(self._get_run_file_path(split, run_idx), features, targets)

    def _validate_split(self, split: str) -> None:
        if split not in self._SPLIT_FOLDERS:
            raise ValueError(f"Unsupported split '{split}' supplied.")

    def _get_run_folders(self, split_path: str) -> Dict[int, str]:
        pattern = self._get_run_folder_pattern()
        content = sorted(os.listdir(split_path))
        run_folders = {int(f[-1]): f for f in content if pattern.match(f) is not None}

        return run_folders

    def _get_run_folder_pattern(self) -> re.Pattern:
        return re.compile(rf"Bearing{self.fd}_\d")

    def _get_scaler_path(self) -> str:
        file_name = f"scaler_{self.fd}_{self.run_split_dist['dev']}.pkl"
        file_path = os.path.join(self._data_root, file_name)

        return file_path

    def _get_run_file_path(self, split: str, run_idx: int) -> str:
        return os.path.join(self._data_root, f"run_{self.fd}_{run_idx}.npy")

    def _get_split_folder(self, split: str) -> str:
        return os.path.join(self._data_root, self._SPLIT_FOLDERS[split])


def _download_femto(data_root: str) -> None:
    with tempfile.TemporaryDirectory() as tmp_path:
        print("Download FEMTO dataset")
        download_path = os.path.join(tmp_path, "FEMTO.zip")
        utils.download_file(FEMTO_URL, download_path)
        print("Extract FEMTO dataset")
        with zipfile.ZipFile(download_path, mode="r") as f:
            f.extractall(data_root)



"""Higher-order data modules to run unsupervised domain adaption experiments."""

import warnings
from copy import deepcopy
from typing import List, Optional, Any, Tuple, Callable, Sequence, Union, cast

import numpy as np
import pytorch_lightning as pl
import torch
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.dataset import ConcatDataset, TensorDataset

from rul_datasets import utils
from rul_datasets.core import PairedRulDataset, RulDataModule


class DomainAdaptionDataModule(pl.LightningDataModule):
    """
    A higher-order [data module][pytorch_lightning.core.LightningDataModule] used for
    unsupervised domain adaption of a labeled source to an unlabeled target domain.
    The training data of both domains is wrapped in a [AdaptionDataset]
    [rul_datasets.adaption.AdaptionDataset] which provides a random sample of the
    target domain with each sample of the source domain. It provides the validation and
    test splits of both domains, and optionally a [paired dataset]
    [rul_datasets.core.PairedRulDataset] for both.

    Examples:
        >>> import rul_datasets
        >>> fd1 = rul_datasets.CmapssReader(fd=1, window_size=20)
        >>> fd2 = rul_datasets.CmapssReader(fd=2, percent_broken=0.8)
        >>> source = rul_datasets.RulDataModule(fd1, 32)
        >>> target = rul_datasets.RulDataModule(fd2, 32)
        >>> dm = rul_datasets.DomainAdaptionDataModule(source, target)
        >>> train_1_2 = dm.train_dataloader()
        >>> val_1, val_2 = dm.val_dataloader()
        >>> test_1, test_2 = dm.test_dataloader()
    """

    def __init__(
        self,
        source: RulDataModule,
        target: RulDataModule,
        paired_val: bool = False,
        inductive: bool = False,
    ) -> None:
        """
        Create a new domain adaption data module from a source and target
        [RulDataModule][rul_datasets.RulDataModule]. The source domain is considered
        labeled and the target domain unlabeled.

        The source and target data modules are checked for compatability (see
        [RulDataModule][rul_datasets.core.RulDataModule.check_compatibility]). These
        checks include that the `fd` differs between them, as they come from the same
        domain otherwise.

        Args:
            source: The data module of the labeled source domain.
            target: The data module of the unlabeled target domain.
            paired_val: Whether to include paired data in validation.
            inductive: Whether to use the target test set for training.
        """
        super().__init__()

        self.source = source
        self.target = target
        self.paired_val = paired_val
        self.batch_size = source.batch_size
        self.inductive = inductive

        self.target_truncated = deepcopy(self.target.reader)
        self.target_truncated.truncate_val = True

        self._check_compatibility()

        self.save_hyperparameters(
            {
                "source": self.source.hparams,
                "target": self.target.hparams,
                "paired_val": self.paired_val,
                "inductive": self.inductive,
            }
        )

    def _check_compatibility(self):
        self.source.check_compatibility(self.target)
        self.target.reader.check_compatibility(self.target_truncated)
        if self.source.reader.fd == self.target.reader.fd:
            raise ValueError(
                f"FD of source and target has to be different for "
                f"domain adaption, but is {self.source.reader.fd} both times."
            )
        if self.target.reader.percent_broken is None:
            warnings.warn(
                "The target domain is not truncated by 'percent_broken'."
                "This may lead to unrealistically good results."
                "If this was intentional, please set `percent_broken` "
                "to 1.0 to silence this warning."
            )

    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        """
        Download and pre-process the underlying data.

        This calls the `prepare_data` function for source and target domain. All
        previously completed preparation steps are skipped. It is called
        automatically by `pytorch_lightning` and executed on the first GPU in
        distributed mode.

        Args:
            *args: Passed down to each data module's `prepare_data` function.
            **kwargs: Passed down to each data module's `prepare_data` function..
        """
        self.source.prepare_data(*args, **kwargs)
        self.target.prepare_data(*args, **kwargs)

    def setup(self, stage: Optional[str] = None) -> None:
        """
        Load source and target domain into memory.

        Args:
            stage: Passed down to each data module's `setup` function.
        """
        self.source.setup(stage)
        self.target.setup(stage)

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """
        Create a data loader of an [AdaptionDataset]
        [rul_datasets.adaption.AdaptionDataset] using source and target domain.

        The data loader is configured to shuffle the data. The `pin_memory` option is
        activated to achieve maximum transfer speed to the GPU.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The training data loader
        """
        return DataLoader(
            self._get_training_dataset(),
            batch_size=self.batch_size,
            shuffle=True,
            pin_memory=True,
        )

    def val_dataloader(self, *args: Any, **kwargs: Any) -> List[DataLoader]:
        """
        Create a data loader of the source, target and paired validation data.

        By default, two data loaders are returned, which correspond to the source
        and the target validation data loader. An optional third is a data loader of a
        [PairedRulDataset][rul_datasets.core.PairedRulDataset] using both source and
        target is returned if `paired_val` was set to `True` in the constructor.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The source, target and an optional paired validation data loader.
        """
        loaders = [
            self.source.val_dataloader(*args, **kwargs),
            self.target.val_dataloader(*args, **kwargs),
        ]
        if self.paired_val:
            loaders.append(
                DataLoader(
                    self._get_paired_dataset(),
                    batch_size=self.batch_size,
                    pin_memory=True,
                )
            )

        return loaders

    def test_dataloader(self, *args: Any, **kwargs: Any) -> List[DataLoader]:
        """
        Create a data loader of the source and target test data.

        The data loaders are the return values of `source.test_dataloader`
        and `target.test_dataloader`.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The source and target test data loader.
        """
        return [
            self.source.test_dataloader(*args, **kwargs),
            self.target.test_dataloader(*args, **kwargs),
        ]

    def _get_training_dataset(self) -> "AdaptionDataset":
        source = self.source.to_dataset("dev")
        target = self.target.to_dataset(
            "test" if self.inductive else "dev", alias="dev"
        )
        dataset = AdaptionDataset(source, target)

        return dataset

    def _get_paired_dataset(self) -> PairedRulDataset:
        paired = PairedRulDataset(
            [self.target_truncated],
            "val",
            num_samples=25000,
            min_distance=1,
            deterministic=True,
        )

        return paired


class LatentAlignDataModule(DomainAdaptionDataModule):
    """
    A higher-order [data module][pytorch_lightning.core.LightningDataModule] based on
    [DomainAdaptionDataModule][rul_datasets.adaption.DomainAdaptionDataModule].

    It is specifically made to work with the latent space alignment approach by Zhang
    et al. The training data of both domains is wrapped in a [AdaptionDataset]
    [rul_datasets.adaption.AdaptionDataset] which splits the data into healthy and
    degrading. For each sample of degrading source data, a random sample of degrading
    target data and healthy sample of either source or target data is drawn. The
    number of steps in degradation are supplied for each degrading sample, as well.
    The data module also provides the validation and test splits of both domains, and
    optionally a [paired dataset][rul_datasets.core.PairedRulDataset] for both.

    Examples:
        >>> import rul_datasets
        >>> fd1 = rul_datasets.CmapssReader(fd=1, window_size=20)
        >>> fd2 = rul_datasets.CmapssReader(fd=2, percent_broken=0.8)
        >>> source = rul_datasets.RulDataModule(fd1, 32)
        >>> target = rul_datasets.RulDataModule(fd2, 32)
        >>> dm = rul_datasets.LatentAlignDataModule(source, target)
        >>> train_1_2 = dm.train_dataloader()
        >>> val_1, val_2 = dm.val_dataloader()
        >>> test_1, test_2 = dm.test_dataloader()
    """

    def __init__(
        self,
        source: RulDataModule,
        target: RulDataModule,
        paired_val: bool = False,
        inductive: bool = False,
        split_by_max_rul: bool = False,
        split_by_steps: Optional[int] = None,
    ) -> None:
        """
        Create a new latent align data module from a source and target
        [RulDataModule][rul_datasets.RulDataModule]. The source domain is considered
        labeled and the target domain unlabeled.

        The source and target data modules are checked for compatability (see
        [RulDataModule][rul_datasets.core.RulDataModule.check_compatibility]). These
        checks include that the `fd` differs between them, as they come from the same
        domain otherwise.

        The healthy and degrading data can be split by either maximum RUL value or
        the number of time steps. See [split_healthy]
        [rul_datasets.adaption.split_healthy] for more information.

        Args:
            source: The data module of the labeled source domain.
            target: The data module of the unlabeled target domain.
            paired_val: Whether to include paired data in validation.
            split_by_max_rul: Whether to split healthy and degrading by max RUL value.
            split_by_steps: Split the healthy and degrading data after this number of
                            time steps.
        """
        super().__init__(source, target, paired_val, inductive)

        if not split_by_max_rul and (split_by_steps is None):
            raise ValueError(
                "Either 'split_by_max_rul' or 'split_by_steps' need to be set."
            )

        self.split_by_max_rul = split_by_max_rul
        self.split_by_steps = split_by_steps

    def _get_training_dataset(self) -> "AdaptionDataset":
        source_healthy, source_degraded = split_healthy(
            *self.source.load_split("dev"), by_max_rul=True
        )
        target_healthy, target_degraded = split_healthy(
            *self.target.load_split("test" if self.inductive else "dev", alias="dev"),
            self.split_by_max_rul,
            self.split_by_steps,
        )
        healthy: Dataset = ConcatDataset([source_healthy, target_healthy])
        dataset = AdaptionDataset(source_degraded, target_degraded, healthy)

        return dataset


def split_healthy(
    features: Union[List[np.ndarray], List[torch.Tensor]],
    targets: Union[List[np.ndarray], List[torch.Tensor]],
    by_max_rul: bool = False,
    by_steps: Optional[int] = None,
) -> Tuple[TensorDataset, TensorDataset]:
    """
    Split the feature and target time series into healthy and degrading parts and
    return a dataset of each.

    If `by_max_rul` is set to `True` the time steps with the maximum RUL value in
    each time series is considered healthy. This option is intended for labeled data
    with piece-wise linear RUL functions. If `by_steps` is set to an integer,
    the first `by_steps` time steps of each series are considered healthy. This
    option is intended for unlabeled data or data with a linear RUL function.

    One option has to be set and both are mutually exclusive.

    Args:
        features: List of feature time series.
        targets: List of target time series.
        by_max_rul: Whether to split healthy and degrading data by max RUL value.
        by_steps: Split healthy and degrading data after this number of time steps.
    Returns:
        healthy: Dataset of healthy data.
        degrading: Dataset of degrading data.
    """
    if not by_max_rul and (by_steps is None):
        raise ValueError("Either 'by_max_rul' or 'by_steps' need to be set.")

    if isinstance(features[0], np.ndarray):
        features, targets = cast(Tuple[List[np.ndarray], ...], (features, targets))
        _features, _targets = utils.to_tensor(features, targets)
    else:
        _features, _targets = cast(Tuple[List[torch.Tensor], ...], (features, targets))

    healthy = []
    degraded = []
    for feature, target in zip(_features, _targets):
        sections = _get_sections(by_max_rul, by_steps, target)
        healthy_feat, degraded_feat = torch.split(feature, sections)
        healthy_target, degraded_target = torch.split(target, sections)
        degradation_steps = torch.arange(1, len(degraded_target) + 1)
        healthy.append((healthy_feat, healthy_target))
        degraded.append((degraded_feat, degradation_steps, degraded_target))

    healthy_dataset = _to_dataset(healthy)
    degraded_dataset = _to_dataset(degraded)

    return healthy_dataset, degraded_dataset


def _get_sections(
    by_max_rul: bool, by_steps: Optional[int], target: torch.Tensor
) -> List[int]:
    # cast is needed for mypy and has no runtime effect
    if by_max_rul:
        split_idx = cast(int, target.flip(0).argmax().item())
        sections = [len(target) - split_idx, split_idx]
    else:
        by_steps = min(cast(int, by_steps), len(target))
        sections = [by_steps, len(target) - by_steps]

    return sections


def _to_dataset(data: Sequence[Tuple[torch.Tensor, ...]]) -> TensorDataset:
    tensor_data = [torch.cat(h) for h in zip(*data)]
    dataset = TensorDataset(*tensor_data)

    return dataset


class AdaptionDataset(Dataset):
    """
    A torch [dataset][torch.utils.data.Dataset] for unsupervised domain adaption. The
    dataset takes a labeled source and one or multiple unlabeled target [dataset]
    [torch.utils.data.Dataset] and combines them.

    For each label/features pair from the source dataset, a random sample of features
    is drawn from each target dataset. The datasets are supposed to provide a sample
    as a tuple of tensors. The target datasets' labels are assumed to be the last
    element of the tuple and are omitted. The datasets length is determined by the
    source dataset. This setup can be used to train with common unsupervised domain
    adaption methods like DAN, DANN or JAN.

    Examples:
        >>> import torch
        >>> import rul_datasets
        >>> source = torch.utils.data.TensorDataset(torch.randn(10), torch.randn(10))
        >>> target = torch.utils.data.TensorDataset(torch.randn(10), torch.randn(10))
        >>> dataset = rul_datasets.adaption.AdaptionDataset(source, target)
        >>> source_features, source_label, target_features = dataset[0]
    """

    _unlabeled_idx: np.ndarray
    _get_unlabeled_idx: Callable

    def __init__(
        self, labeled: Dataset, *unlabeled: Dataset, deterministic: bool = False
    ) -> None:
        """
        Create a new adaption data set from a labeled source and one or multiple
        unlabeled target dataset.

        By default, a random sample is drawn from each target dataset when a source
        sample is accessed. This is the recommended setting for training. To
        deactivate this behavior and fix the pairing of source and target samples,
        set `deterministic` to `True`. This is the recommended setting for evaluation.

        Args:
            labeled: The dataset from the labeled domain.
            *unlabeled: The dataset(s) from the unlabeled domain(s).
            deterministic: Return the same target sample for each source sample.
        """
        self.labeled = labeled
        self.unlabeled = unlabeled
        self.deterministic = deterministic
        self._unlabeled_len = [len(ul) for ul in self.unlabeled]  # type: ignore

        if self.deterministic:
            self._rng = np.random.default_rng(seed=42)
            size = (len(self), len(self._unlabeled_len))
            self._unlabeled_idx = self._rng.integers(0, self._unlabeled_len, size)
            self._get_unlabeled_idx = self._get_deterministic_unlabeled_idx
        else:
            self._rng = np.random.default_rng()
            self._get_unlabeled_idx = self._get_random_unlabeled_idx

    def _get_random_unlabeled_idx(self, _: int) -> np.ndarray:
        return self._rng.integers(0, self._unlabeled_len)

    def _get_deterministic_unlabeled_idx(self, idx: int) -> np.ndarray:
        return self._unlabeled_idx[idx]

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, ...]:
        item = self.labeled[idx]
        for unlabeled, ul_idx in zip(self.unlabeled, self._get_unlabeled_idx(idx)):
            item += unlabeled[ul_idx][:-1]  # drop label tensor in last position

        return item

    def __len__(self) -> int:
        return len(self.labeled)  # type: ignore


class PretrainingAdaptionDataModule(pl.LightningDataModule):
    def __init__(
        self,
        source: RulDataModule,
        target: RulDataModule,
        num_samples: int,
        min_distance: int = 1,
        distance_mode: str = "linear",
    ):
        super().__init__()

        self.source = source
        self.target = target
        self.num_samples = num_samples
        self.batch_size = source.batch_size
        self.min_distance = min_distance
        self.distance_mode = distance_mode

        self.target_loader = self.target.reader
        self.source_loader = self.source.reader

        self._check_compatibility()

        self.save_hyperparameters(
            {
                "fd_source": self.source_loader.fd,
                "fd_target": self.target_loader.fd,
                "num_samples": self.num_samples,
                "batch_size": self.batch_size,
                "window_size": self.source_loader.window_size,
                "max_rul": self.source_loader.max_rul,
                "min_distance": self.min_distance,
                "percent_broken": self.target_loader.percent_broken,
                "percent_fail_runs": self.target_loader.percent_fail_runs,
                "truncate_target_val": self.target_loader.truncate_val,
                "distance_mode": self.distance_mode,
            }
        )

    def _check_compatibility(self):
        self.source.check_compatibility(self.target)
        if self.source_loader.fd == self.target_loader.fd:
            raise ValueError(
                f"FD of source and target has to be different for "
                f"domain adaption, but is {self.source_loader.fd} bot times."
            )
        if (
            self.target_loader.percent_broken is None
            or self.target_loader.percent_broken == 1.0
        ):
            raise ValueError(
                "Target data needs a percent_broken smaller than 1 for pre-training."
            )
        if (
            self.source_loader.percent_broken is not None
            and self.source_loader.percent_broken < 1.0
        ):
            raise ValueError(
                "Source data cannot have a percent_broken smaller than 1, "
                "otherwise it would not be failed, labeled data."
            )
        if not self.target_loader.truncate_val:
            warnings.warn(
                "Validation data of unfailed runs is not truncated. "
                "The validation metrics will not be valid."
            )

    def prepare_data(self, *args, **kwargs):
        self.source_loader.prepare_data()
        self.target_loader.prepare_data()

    def setup(self, stage: Optional[str] = None):
        self.source.setup(stage)
        self.target.setup(stage)

    def train_dataloader(self, *args, **kwargs) -> DataLoader:
        return DataLoader(
            self._get_paired_dataset("dev"), batch_size=self.batch_size, pin_memory=True
        )

    def val_dataloader(self, *args, **kwargs) -> List[DataLoader]:
        combined_loader = DataLoader(
            self._get_paired_dataset("val"), batch_size=self.batch_size, pin_memory=True
        )
        source_loader = self.source.val_dataloader()
        target_loader = self.target.val_dataloader()

        return [combined_loader, source_loader, target_loader]

    def _get_paired_dataset(self, split: str) -> PairedRulDataset:
        deterministic = split == "val"
        min_distance = 1 if split == "val" else self.min_distance
        num_samples = 50000 if split == "val" else self.num_samples
        paired = PairedRulDataset(
            [self.source_loader, self.target_loader],
            split,
            num_samples,
            min_distance,
            deterministic,
            mode=self.distance_mode,
        )

        return paired



"""This module contains the base class for all readers. It is only relevant to people
that want to extend this package with their own dataset. """
import abc
from copy import deepcopy
from typing import Optional, Union, List, Dict, Any, Iterable, Tuple, Literal

import numpy as np

from rul_datasets.reader import truncating


class AbstractReader(metaclass=abc.ABCMeta):
    """
    This reader is the abstract base class of all readers.

    In case you want to extend this library with a dataset of your own, you should
    create a subclass of `AbstractReader`. It defines the public interface that all
    data modules in this library use. Just inherit from this class implement the
    abstract functions, and you should be good to go.

    Please consider contributing your work afterward to help the community.

    Examples:
        >>> import rul_datasets
        >>> class MyReader(rul_datasets.reader.AbstractReader):
        ...     def fds(self):
        ...         return [1]
        ...
        ...     def prepare_data(self):
        ...         pass
        ...
        ...     def default_window_size(self, fd):
        ...         return 30
        ...
        ...     def load_complete_split(self, split, alias):
        ...         features = [np.random.randn(100, 2, 30) for _ in range(10)]
        ...         targets = [np.arange(100, 0, -1) for _ in range(10)]
        ...
        ...         return features, targets
        ...
        >>> my_reader = MyReader(fd=1)
        >>> features, targets = my_reader.load_split("dev")
        >>> features[0].shape
        torch.Size([100, 2, 30])
    """

    fd: int
    window_size: int
    max_rul: Optional[int]
    percent_broken: Optional[float]
    percent_fail_runs: Optional[Union[float, List[int], None]]
    truncate_val: bool
    truncate_degraded_only: bool

    _NUM_TRAIN_RUNS: Dict[int, int]

    def __init__(
        self,
        fd: int,
        window_size: Optional[int] = None,
        max_rul: Optional[int] = None,
        percent_broken: Optional[float] = None,
        percent_fail_runs: Optional[Union[float, List[int]]] = None,
        truncate_val: bool = False,
        truncate_degraded_only: bool = False,
    ) -> None:
        """
        Create a new reader. If your reader needs additional input arguments,
        create your own `__init__` function and call this one from within as `super(
        ).__init__(...)`.

        For more information about using readers refer to the [reader]
        [rul_datasets.reader] module page.

        Args:
            fd: Index of the selected sub-dataset
            window_size: Size of the sliding window. Defaults to 2560.
            max_rul: Maximum RUL value of targets.
            percent_broken: The maximum relative degradation per time series.
            percent_fail_runs: The percentage or index list of available time series.
            truncate_val: Truncate the validation data with `percent_broken`, too.
            truncate_degraded_only: Only truncate the degraded part of the data
                                    (< max RUL).
        """
        self.fd = fd
        self.window_size = window_size or self.default_window_size(self.fd)
        self.max_rul = max_rul
        self.truncate_val = truncate_val
        self.percent_broken = percent_broken
        self.percent_fail_runs = percent_fail_runs
        self.truncate_degraded_only = truncate_degraded_only

    @property
    def hparams(self) -> Dict[str, Any]:
        """All information logged by the data modules as hyperparameters in PyTorch
        Lightning."""
        return {
            "dataset": self.dataset_name,
            "fd": self.fd,
            "window_size": self.window_size,
            "max_rul": self.max_rul,
            "percent_broken": self.percent_broken,
            "percent_fail_runs": self.percent_fail_runs,
            "truncate_val": self.truncate_val,
            "truncate_degraded_only": self.truncate_degraded_only,
        }

    @property
    @abc.abstractmethod
    def dataset_name(self) -> str:
        """Name of the dataset."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def fds(self) -> List[int]:
        """The indices of available sub-datasets."""
        raise NotImplementedError

    @abc.abstractmethod
    def prepare_data(self) -> None:
        """Prepare the data. This function should take care of things that need to be
        done once, before the data can be used. This may include downloading,
        extracting or transforming the data, as well as fitting scalers. It is best
        practice to check if a preparation step was completed before to avoid
        repeating it unnecessarily."""
        raise NotImplementedError

    @abc.abstractmethod
    def default_window_size(self, fd: int) -> int:
        """
        The default window size of the data set. This may vary from sub-dataset to
        sub-dataset.

        Args:
            fd: The index of a sub-dataset.
        Returns:
            The default window size for the sub-dataset.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def load_complete_split(
        self, split: str, alias: str
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        Load a complete split without truncation.

        This function should return the features and targets of the desired split.
        Both should be contained in a list of numpy arrays. Each of the arrays
        contains one time series. The features should have a shape of `[num_windows,
        window_size, num_channels]` and the targets a shape of `[num_windows]`. The
        features should be scaled as desired. The targets should be capped by
        `max_rul`.

        By setting `alias`, it should be possible to load a split aliased as another
        split, e.g. load the test split and treat it as the dev split. The data of
        `split` should be loaded but all pre-processing steps of `alias` should be
        carried out.

        This function is used internally in [load_split]
        [rul_datasets.reader.abstract.AbstractReader.load_split] which takes care of
        truncation.

        Args:
            split: The name of the split to load.
            alias: The split as which the loaded data should be treated.
        Returns:
            features: The complete, scaled features of the desired split.
            targets: The capped target values corresponding to the features.
        """
        raise NotImplementedError

    def load_split(
        self, split: str, alias: Optional[str] = None
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        """
        Load a split as tensors and apply truncation to it.

        This function loads the scaled features and the targets of a split into
        memory. Afterwards, truncation is applied if the `split` is set to `dev`. The
        validation set is also truncated with `percent_broken` if `truncate_val` is
        set to `True`.

        By setting `alias`, it is possible to load a split aliased as another split,
        e.g. load the test split and treat it as the dev split. The data of `split`
        is loaded but all pre-processing steps of `alias` are carried out.

        Args:
            split: The desired split to load.
            alias: The split as which the loaded data should be treated.
        Returns:
            features: The scaled, truncated features of the desired split.
            targets: The truncated targets of the desired split.
        """
        alias = alias or split
        features, targets = self.load_complete_split(split, alias)
        if alias == "dev":
            features, targets = truncating.truncate_runs(
                features,
                targets,
                self.percent_broken,
                self.percent_fail_runs,
                self.truncate_degraded_only,
            )
        elif alias == "val" and self.truncate_val:
            features, targets = truncating.truncate_runs(
                features,
                targets,
                self.percent_broken,
                degraded_only=self.truncate_degraded_only,
            )

        return features, targets

    def get_compatible(
        self,
        fd: Optional[int] = None,
        percent_broken: Optional[float] = None,
        percent_fail_runs: Union[float, List[int], None] = None,
        truncate_val: Optional[bool] = None,
        consolidate_window_size: Literal["override", "min", "none"] = "override",
    ) -> "AbstractReader":
        """
        Create a new reader of the desired sub-dataset that is compatible to this one
        (see [check_compatibility]
        [rul_datasets.reader.abstract.AbstractReader.check_compatibility]). Useful for
        domain adaption.

        The values for `percent_broken`, `percent_fail_runs` and `truncate_val` of
        the new reader can be overridden.

        When constructing a compatible reader for another sub-dataset, the window
        size of this reader will be used to override the default window size of the
        new reader. This behavior can be changed by setting `consolidate_window_size`
        to `"min"`. The window size of this reader and the new one will be set to the
        minimum of this readers window size and the default window size of the new
        reader. Please be aware that this will change the window size of *this*
        reader, too. If the new reader should use its default window size,
        set `consolidate_window_size` to `"none"`.

        Args:
            fd: The index of the sub-dataset for the new reader.
            percent_broken: Override this value in the new reader.
            percent_fail_runs: Override this value in the new reader.
            truncate_val: Override this value in the new reader.
            consolidate_window_size: How to consolidate the window size of the readers.
        Returns:
            A compatible reader with optional overrides.
        """
        other = deepcopy(self)
        if percent_broken is not None:
            other.percent_broken = percent_broken
        if percent_fail_runs is not None:
            other.percent_fail_runs = percent_fail_runs
        if truncate_val is not None:
            other.truncate_val = truncate_val
        if fd is not None:
            other.fd = fd
        self._consolidate_window_size(other, consolidate_window_size)

        return other

    def _consolidate_window_size(
        self, other: "AbstractReader", mode: Literal["override", "min", "none"]
    ) -> None:
        if mode == "override":
            other.window_size = self.window_size
        elif mode == "min":
            window_size = min(self.window_size, self.default_window_size(other.fd))
            self.window_size = window_size
            other.window_size = window_size
        elif mode == "none":
            other.window_size = self.default_window_size(other.fd)

    def get_complement(
        self,
        percent_broken: Optional[float] = None,
        truncate_val: Optional[bool] = None,
    ) -> "AbstractReader":
        """
        Get a compatible reader that contains all development runs that are not in
        this reader (see [check_compatibility]
        [rul_datasets.reader.abstract.AbstractReader.check_compatibility]). Useful for
        semi-supervised learning.

        The new reader will contain the development runs that were discarded in this
        reader due to truncation through `percent_fail_runs`. If `percent_fail_runs`
        was not set or this reader contains all development runs, it returns a reader
        with an empty development set.

        The values for `percent_broken`, and `truncate_val` of the new reader can be
        overridden.

        Args:
            percent_broken: Override this value in the new reader.
            truncate_val: Override this value in the new reader.
        Returns:
            A compatible reader with all development runs missing in this one.
        """
        complement_idx = self._get_complement_idx()
        other = self.get_compatible(
            percent_broken=percent_broken,
            percent_fail_runs=complement_idx,
            truncate_val=truncate_val,
        )

        return other

    def _get_complement_idx(self) -> List[int]:
        num_runs = self._NUM_TRAIN_RUNS[self.fd]
        run_idx = list(range(num_runs))
        if isinstance(self.percent_fail_runs, float):
            split_idx = int(self.percent_fail_runs * num_runs)
            complement_idx = run_idx[split_idx:]
        elif isinstance(self.percent_fail_runs, Iterable):
            complement_idx = list(set(run_idx).difference(self.percent_fail_runs))
        else:
            complement_idx = []

        return complement_idx

    def is_mutually_exclusive(self, other: "AbstractReader") -> bool:
        """
        Check if this reader is mutually exclusive to another reader.

        Two readers are mutually exclusive if:

        * they are not of the same class and therefore do not share a dataset
        * their `percent_fail_runs` arguments do not overlap (float arguments overlap
          if they are greater than zero)
        * one of them is empty

        Args:
            other: The reader to check exclusivity against.
        Returns:
            Whether the readers are mutually exclusive.
        """
        self_runs = 1.0 if self.percent_fail_runs is None else self.percent_fail_runs
        other_runs = 1.0 if other.percent_fail_runs is None else other.percent_fail_runs

        if not isinstance(self, type(other)):
            mutually_exclusive = True
        elif self_runs == other and self_runs and other_runs:
            mutually_exclusive = False  # both the same and not empty
        elif isinstance(self_runs, float) and isinstance(other_runs, float):
            mutually_exclusive = False  # both start with first run -> overlap
        elif isinstance(self_runs, float) and isinstance(other_runs, Iterable):
            mutually_exclusive = self._is_mutually_exclusive(self, other)
        elif isinstance(self_runs, Iterable) and isinstance(other_runs, float):
            mutually_exclusive = self._is_mutually_exclusive(other, self)
        else:
            mutually_exclusive = set(self_runs).isdisjoint(other_runs)  # type: ignore

        return mutually_exclusive

    def _is_mutually_exclusive(
        self, floated: "AbstractReader", listed: "AbstractReader"
    ) -> bool:
        """Listed is mutually exclusive if it is a subset of floated's complement."""
        floated_complement = floated.get_complement().percent_fail_runs
        listed = listed.percent_fail_runs  # type: ignore
        exclusive = set(listed).issubset(floated_complement)  # type: ignore

        return exclusive

    def check_compatibility(self, other: "AbstractReader") -> None:
        """
        Check if the other reader is compatible with this one.

        Compatibility of two readers ensures that training with both will probably
        succeed and produce valid results. Two readers are considered compatible, if
        they:

        * are both children of [AbstractReader]
        [rul_datasets.reader.abstract.AbstractReader]

        * have the same `window size`

        * have the same `max_rul`

        If any of these conditions is not met, the readers are considered
        misconfigured and a `ValueError` is thrown.

        Args:
            other: Another reader object.
        """
        if not isinstance(other, type(self)):
            raise ValueError(
                f"The other loader is not of class {type(self)} but {type(other)}."
            )
        if not self.window_size == other.window_size:
            raise ValueError(
                f"Window sizes are not compatible "
                f"{self.window_size} vs. {other.window_size}"
            )
        if not self.max_rul == other.max_rul:
            raise ValueError(
                f"Max RULs are not compatible " f"{self.max_rul} vs. {other.max_rul}"
            )



import os
from typing import List, Optional, Callable, Dict, Tuple

import numpy as np
import requests  # type: ignore
import torch
from tqdm import tqdm  # type: ignore


def get_files_in_path(path: str, condition: Optional[Callable] = None) -> List[str]:
    """
    Return the paths of all files in a path that satisfy a condition in alphabetical
    order.

    If the condition is `None` all files are returned.

    Args:
        path: the path to look into
        condition: the include-condition for files
    Returns:
        all files that satisfy the condition in alphabetical order
    """
    if condition is None:
        feature_files = [f for f in os.listdir(path)]
    else:
        feature_files = [f for f in os.listdir(path) if condition(f)]
    feature_files = sorted(os.path.join(path, f) for f in feature_files)

    return feature_files


def get_targets_from_file_paths(
    file_paths: Dict[int, List[str]], timestep_from_file_path: Callable
) -> Dict[int, np.ndarray]:
    """
    Create the RUL targets based on the file paths of the feature files.

    The function extracts the feature file path from each path. The supplied
    conversion function extracts the time step from it. Afterwards the RUL is
    calculated by subtracting each time step from the maximum time step plus 1.

    Args:
        file_paths: runs represented as dict of feature file paths
        timestep_from_file_path: Function to convert a feature file path to a time step

    Returns:
        A list of RUL target arrays for each run
    """
    targets = {}
    for run_idx, run_files in file_paths.items():
        run_targets = np.empty(len(run_files))
        for i, file_path in enumerate(run_files):
            run_targets[i] = timestep_from_file_path(file_path)
        run_targets = np.max(run_targets) - run_targets + 1
        targets[run_idx] = run_targets

    return targets


def extract_windows(seq: np.ndarray, window_size: int) -> np.ndarray:
    """
    Extract sliding windows from a sequence.

    The step size is considered to be one, which results in `len(seq) - window_size +
    1` extracted windows. The resulting array has the shape [num_windows, window_size,
    num_channels].

    Args:
        seq: sequence to extract windows from
        window_size: length of the sliding window
    Returns:
        array of sliding windows
    """
    if window_size > len(seq):
        raise ValueError(
            f"Cannot extract windows of size {window_size} "
            f"from a sequence of length {len(seq)}."
        )

    num_frames = seq.shape[0] - window_size + 1
    window_idx = np.arange(window_size)[None, :] + np.arange(num_frames)[:, None]
    windows = seq[window_idx]

    return windows


def download_file(url: str, save_path: str) -> None:
    response = requests.get(url, stream=True)
    if not response.status_code == 200:
        raise RuntimeError(f"Download failed. Server returned {response.status_code}")
    content_len = int(response.headers["Content-Length"]) // 1024
    with open(save_path, mode="wb") as f:
        for data in tqdm(response.iter_content(chunk_size=1024), total=content_len):
            f.write(data)


def to_tensor(
    features: List[np.ndarray], *targets: List[np.ndarray]
) -> Tuple[List[torch.Tensor], ...]:
    dtype = torch.float32
    tensor_feats = [feature_to_tensor(f, dtype) for f in features]
    tensor_targets = [
        [torch.tensor(t, dtype=dtype) for t in target] for target in targets
    ]

    return tensor_feats, *tensor_targets


def feature_to_tensor(features: np.ndarray, dtype: torch.dtype) -> torch.Tensor:
    if len(features.shape) == 2:
        return torch.tensor(features, dtype=dtype).permute(1, 0)
    else:
        return torch.tensor(features, dtype=dtype).permute(0, 2, 1)



"""Higher-order data modules to establish a baseline for transfer learning and domain
adaption experiments. """

import warnings
from copy import deepcopy
from typing import List, Optional, Any

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from rul_datasets.core import PairedRulDataset, RulDataModule


class BaselineDataModule(pl.LightningDataModule):
    """
    A higher-order [data module][pytorch_lightning.core.LightningDataModule] that
    takes a [RulDataModule][rul_datasets.core.RulDataModule]. It provides the
    training and validation splits of the sub-dataset selected in the underlying data
    module but provides the test splits of all available subsets of the dataset. This
    makes it easy to evaluate the generalization of a supervised model on all
    sub-datasets.

    Examples:
        >>> import rul_datasets
        >>> cmapss = rul_datasets.reader.CmapssReader(fd=1)
        >>> dm = rul_datasets.RulDataModule(cmapss, batch_size=32)
        >>> baseline_dm = rul_datasets.BaselineDataModule(dm)
        >>> train_fd1 = baseline_dm.train_dataloader()
        >>> val_fd1 = baseline_dm.val_dataloader()
        >>> test_fd1, test_fd2, test_fd3, test_fd4 = baseline_dm.test_dataloader()

    """

    def __init__(self, data_module: RulDataModule) -> None:
        """
        Create a new baseline data module from a [RulDataModule]
        [rul_datasets.RulDataModule].

        It will provide a data loader of the underlying data module's training and
        validation splits. Additionally, it provides a data loader of the test split
        of all sub-datasets.

        The data module keeps the configuration made in the underlying data module.
        The same configuration is then passed on to create RulDataModules for all
        sub-datasets, beside `percent_fail_runs` and `percent_broken`.

        Args:
            data_module: the underlying RulDataModule
        """
        super().__init__()

        self.data_module = data_module
        self.save_hyperparameters(self.data_module.hparams)

        self.subsets = {}
        for fd in self.data_module.fds:
            self.subsets[fd] = self._get_fd(fd)

    def _get_fd(self, fd):
        if fd == self.data_module.reader.fd:
            dm = self.data_module
        else:
            loader = deepcopy(self.data_module.reader)
            loader.fd = fd
            loader.percent_fail_runs = None
            loader.percent_broken = None
            dm = RulDataModule(loader, self.data_module.batch_size)

        return dm

    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        """
        Download and pre-process the underlying data.

        This calls the `prepare_data` function for all sub-datasets. All
        previously completed preparation steps are skipped. It is called
        automatically by `pytorch_lightning` and executed on the first GPU in
        distributed mode.

        Args:
            *args: Passed down to each data module's `prepare_data` function.
            **kwargs: Passed down to each data module's `prepare_data` function..
        """
        for dm in self.subsets.values():
            dm.prepare_data(*args, **kwargs)

    def setup(self, stage: Optional[str] = None):
        """
        Load all splits as tensors into memory.

        Args:
            stage: Passed down to each data module's `setup` function.
        """
        for dm in self.subsets.values():
            dm.setup(stage)

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """See [rul_datasets.core.RulDataModule.train_dataloader][]."""
        return self.data_module.train_dataloader(*args, **kwargs)

    def val_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """See [rul_datasets.core.RulDataModule.val_dataloader][]."""
        return self.data_module.val_dataloader(*args, **kwargs)

    def test_dataloader(self, *args: Any, **kwargs: Any) -> List[DataLoader]:
        """
        Return data loaders for all sub-datasets.

        Args:
            *args: Passed down to each data module.
            **kwargs: Passed down to each data module.
        Returns:
            The test dataloaders of all sub-datasets.
        """
        test_dataloaders = []
        for fd_target in self.data_module.fds:
            target_dl = self.subsets[fd_target].test_dataloader(*args, **kwargs)
            test_dataloaders.append(target_dl)

        return test_dataloaders


class PretrainingBaselineDataModule(pl.LightningDataModule):
    def __init__(
        self,
        failed_data_module: RulDataModule,
        unfailed_data_module: RulDataModule,
        num_samples: int,
        min_distance: int = 1,
        distance_mode: str = "linear",
    ):
        super().__init__()

        self.failed_loader = failed_data_module.reader
        self.unfailed_loader = unfailed_data_module.reader
        self.num_samples = num_samples
        self.batch_size = failed_data_module.batch_size
        self.min_distance = min_distance
        self.distance_mode = distance_mode
        self.window_size = self.unfailed_loader.window_size
        self.source = unfailed_data_module

        self._check_loaders()

        self.save_hyperparameters(
            {
                "fd_source": self.unfailed_loader.fd,
                "num_samples": self.num_samples,
                "batch_size": self.batch_size,
                "window_size": self.window_size,
                "max_rul": self.unfailed_loader.max_rul,
                "min_distance": self.min_distance,
                "percent_broken": self.unfailed_loader.percent_broken,
                "percent_fail_runs": self.failed_loader.percent_fail_runs,
                "truncate_val": self.unfailed_loader.truncate_val,
                "distance_mode": self.distance_mode,
            }
        )

    def _check_loaders(self):
        self.failed_loader.check_compatibility(self.unfailed_loader)
        if not self.failed_loader.fd == self.unfailed_loader.fd:
            raise ValueError("Failed and unfailed data need to come from the same FD.")
        if self.failed_loader.percent_fail_runs is None or isinstance(
            self.failed_loader.percent_fail_runs, float
        ):
            raise ValueError(
                "Failed data needs list of failed runs "
                "for pre-training but uses a float or is None."
            )
        if self.unfailed_loader.percent_fail_runs is None or isinstance(
            self.unfailed_loader.percent_fail_runs, float
        ):
            raise ValueError(
                "Unfailed data needs list of failed runs "
                "for pre-training but uses a float or is None."
            )
        if set(self.failed_loader.percent_fail_runs).intersection(
            self.unfailed_loader.percent_fail_runs
        ):
            raise ValueError(
                "Runs of failed and unfailed data overlap. "
                "Please use mututally exclusive sets of runs."
            )
        if (
            self.unfailed_loader.percent_broken is None
            or self.unfailed_loader.percent_broken == 1.0
        ):
            raise ValueError(
                "Unfailed data needs a percent_broken smaller than 1 for pre-training."
            )
        if (
            self.failed_loader.percent_broken is not None
            and self.failed_loader.percent_broken < 1.0
        ):
            raise ValueError(
                "Failed data cannot have a percent_broken smaller than 1, "
                "otherwise it would not be failed data."
            )
        if not self.unfailed_loader.truncate_val:
            warnings.warn(
                "Validation data of unfailed runs is not truncated. "
                "The validation metrics will not be valid."
            )

    def prepare_data(self, *args, **kwargs):
        self.unfailed_loader.prepare_data()

    def setup(self, stage: Optional[str] = None):
        self.source.setup(stage)

    def train_dataloader(self, *args, **kwargs) -> DataLoader:
        return DataLoader(
            self._get_paired_dataset("dev"), batch_size=self.batch_size, pin_memory=True
        )

    def val_dataloader(self, *args, **kwargs) -> List[DataLoader]:
        combined_loader = DataLoader(
            self._get_paired_dataset("val"), batch_size=self.batch_size, pin_memory=True
        )
        source_loader = self.source.val_dataloader()

        return [combined_loader, source_loader]

    def _get_paired_dataset(self, split: str) -> PairedRulDataset:
        deterministic = split == "val"
        min_distance = 1 if split == "val" else self.min_distance
        num_samples = 25000 if split == "val" else self.num_samples
        paired = PairedRulDataset(
            [self.unfailed_loader, self.failed_loader],
            split,
            num_samples,
            min_distance,
            deterministic,
            mode=self.distance_mode,
        )

        return paired



"""A module for working with the data root directory."""

import os


def _init_data_root() -> str:
    if "RUL_DATASETS_DATA_ROOT" in os.environ:
        root = os.environ["RUL_DATASETS_DATA_ROOT"]
        if not os.path.exists(root):
            raise ValueError(
                f"Data root '{root}' set by 'RUL_DATASETS_DATA_ROOT' "
                "env var does not exist."
            )
    else:
        root = os.path.expanduser(os.path.join("~", ".rul-datasets"))
        os.makedirs(root, exist_ok=True)

    return root


_DATA_ROOT = _init_data_root()


def get_data_root() -> str:
    """
    Return the path to the data root directory.

    The default data root is located at `~/.rul-datasets`. You can customize the
    location by setting the environment variable `RUL_DATASETS_DATA_ROOT` or by
    calling [set_data_root][rul_datasets.reader.data_root.set_data_root].

    A manually set data root must be an already existing directory.

    Returns:
        The data root path.
    """
    return _DATA_ROOT


def set_data_root(data_root: str) -> None:
    """
    Set the data root to the specified location.

    A manually set data root must be an already existing directory.

    Args:
        data_root: The new data root location.
    """
    global _DATA_ROOT
    if not os.path.exists(data_root):
        raise ValueError(f"Data root '{data_root}' does not exist.")
    _DATA_ROOT = data_root



"""A module with higher-order data modules for semi-supervised learning."""
import warnings
from typing import Any, Optional

import pytorch_lightning as pl
from torch.utils.data import DataLoader

from rul_datasets.adaption import AdaptionDataset
from rul_datasets.core import RulDataModule


class SemiSupervisedDataModule(pl.LightningDataModule):
    """
    A higher-order [data module][pytorch_lightning.core.LightningDataModule] used for
    semi-supervised learning with a labeled data module and an unlabeled one. It
    makes sure that both data modules come from the same sub-dataset.

    Examples:
        >>> import rul_datasets
        >>> fd1 = rul_datasets.CmapssReader(fd=1, window_size=20, percent_fail_runs=0.5)
        >>> fd1_complement = fd1.get_complement(percent_broken=0.8)
        >>> labeled = rul_datasets.RulDataModule(fd1, 32)
        >>> unlabeled = rul_datasets.RulDataModule(fd1_complement, 32)
        >>> dm = rul_datasets.SemiSupervisedDataModule(labeled, unlabeled)
        >>> train_ssl = dm.train_dataloader()
        >>> val = dm.val_dataloader()
        >>> test = dm.test_dataloader()
    """

    def __init__(self, labeled: RulDataModule, unlabeled: RulDataModule) -> None:
        """
        Create a new semi-supervised data module from a labeled and unlabeled
        [RulDataModule][rul_datasets.RulDataModule].

        The both data modules are checked for compatability (see[RulDataModule]
        [rul_datasets.core.RulDataModule.check_compatibility]). These
        checks include that the `fd` match between them.

        Args:
            labeled: The data module of the labeled dataset.
            unlabeled: The data module of the unlabeled dataset.
        """
        super().__init__()

        self.labeled = labeled
        self.unlabeled = unlabeled
        self.batch_size = labeled.batch_size

        self._check_compatibility()

        self.save_hyperparameters(
            {"labeled": self.labeled.hparams, "unlabeled": self.unlabeled.hparams}
        )

    def _check_compatibility(self) -> None:
        self.labeled.check_compatibility(self.unlabeled)
        if not self.labeled.reader.fd == self.unlabeled.reader.fd:
            raise ValueError(
                "FD of source and target has to be the same for "
                "semi-supervised learning, but they are "
                f"{self.labeled.reader.fd} and {self.unlabeled.reader.fd}."
            )
        if self.unlabeled.reader.percent_broken is None:
            warnings.warn(
                "The unlabeled data is not truncated by 'percent_broken'."
                "This may lead to unrealistically good results."
                "If this was intentional, please set `percent_broken` "
                "to 1.0 to silence this warning."
            )
        if not self.labeled.is_mutually_exclusive(self.unlabeled):
            warnings.warn(
                "The data modules are not mutually exclusive. "
                "This means there is an overlap between labeled and "
                "unlabeled data, which should not be that case for "
                "semi-supervised learning. You can check this by calling "
                "'is_mutually_exclusive' on a reader or RulDataModule."
            )

    def prepare_data(self, *args: Any, **kwargs: Any) -> None:
        """
        Download and pre-process the underlying data.

        This calls the `prepare_data` function for source and target domain. All
        previously completed preparation steps are skipped. It is called
        automatically by `pytorch_lightning` and executed on the first GPU in
        distributed mode.

        Args:
            *args: Passed down to each data module's `prepare_data` function.
            **kwargs: Passed down to each data module's `prepare_data` function..
        """
        self.labeled.prepare_data(*args, **kwargs)
        self.unlabeled.prepare_data(*args, **kwargs)

    def setup(self, stage: Optional[str] = None) -> None:
        """
        Load labeled and unlabeled data into memory.

        Args:
            stage: Passed down to each data module's `setup` function.
        """
        self.labeled.setup(stage)
        self.unlabeled.setup(stage)

    def train_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """
        Create a data loader of an [AdaptionDataset]
        [rul_datasets.adaption.AdaptionDataset] using labeled and unlabeled.

        The data loader is configured to shuffle the data. The `pin_memory` option is
        activated to achieve maximum transfer speed to the GPU.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The training data loader
        """
        return DataLoader(
            self._to_dataset("dev"),
            batch_size=self.batch_size,
            shuffle=True,
            pin_memory=True,
        )

    def val_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """
        Create a data loader of the labeled validation data.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The labeled validation data loader.
        """
        return self.labeled.val_dataloader()

    def test_dataloader(self, *args: Any, **kwargs: Any) -> DataLoader:
        """
        Create a data loader of the labeled test data.

        Args:
            *args: Ignored. Only for adhering to parent class interface.
            **kwargs: Ignored. Only for adhering to parent class interface.
        Returns:
            The labeled test data loader.
        """
        return self.labeled.test_dataloader()

    def _to_dataset(self, split: str) -> "AdaptionDataset":
        labeled = self.labeled.to_dataset(split)
        unlabeled = self.unlabeled.to_dataset(split)
        dataset = AdaptionDataset(labeled, unlabeled)

        return dataset



"""A module with functions for efficient saving and loading of RUL features and
targets. """

import os.path
from typing import Tuple, List, Dict, Literal, Optional

import numpy as np
from tqdm import tqdm  # type: ignore


def save(save_path: str, features: np.ndarray, targets: np.ndarray) -> None:
    """
    Save features and targets of a run to .npy files.

    The arrays are saved to separate .npy files to enable memmap mode in case RAM is
    short. The files are saved as <save_path>_features.npy and
    <save_path>_targets.npy. To load the files, supply the same `save_path` to the
    [load][rul_datasets.reader.saving.load] function. If the `save_path` does not have
    the .npy file extension .npy will be appended.

    Args:
          save_path: The path including file name to save the arrays to.
          features: The feature array to save.
          targets: The targets array to save.
    """
    feature_path = _get_feature_path(save_path)
    np.save(feature_path, features, allow_pickle=False)
    target_path = _get_target_path(save_path)
    np.save(target_path, targets, allow_pickle=False)


def load(save_path: str, memmap: bool = False) -> Tuple[np.ndarray, np.ndarray]:
    """
    Load features and targets of a run from .npy files.

    This method is used to restore runs that were saved with the [save]
    [rul_datasets.reader.saving.save] function. If the runs are too large for the RAM,
    `memmap` can be set to True to avoid reading them completely to memory. This
    results in slower processing, though.

    Args:
        save_path: Path that was supplied to the
                   [save][rul_datasets.reader.saving.save] function.
        memmap: whether to use memmap to avoid loading the whole run into memory
    Returns:
        features: The feature array saved in `save_path`
        targets: The target array saved in `save_path`
    """
    memmap_mode: Optional[Literal["r"]] = "r" if memmap else None
    feature_path = _get_feature_path(save_path)
    features = np.load(feature_path, memmap_mode, allow_pickle=False)
    target_path = _get_target_path(save_path)
    targets = np.load(target_path, memmap_mode, allow_pickle=False)

    return features, targets


def load_multiple(
    save_paths: List[str], memmap: bool = False
) -> Tuple[List[np.ndarray], List[np.ndarray]]:
    """
    Load multiple runs with the [load][rul_datasets.reader.saving.load] function.

    Args:
        save_paths: The list of run files to load.
        memmap: See [load][rul_datasets.reader.saving.load]
    Returns:
        features: The feature arrays saved in `save_paths`
        targets: The target arrays saved in `save_paths`
    """
    if save_paths:
        runs = [load(save_path, memmap) for save_path in save_paths]
        features, targets = [list(x) for x in zip(*runs)]
    else:
        features, targets = [], []

    return features, targets


def exists(save_path: str) -> bool:
    """
    Return if the files resulting from a `save` call with `save_path` exist.

    Args:
        save_path: the `save_path` the [save][rul_datasets.reader.saving.save]
                   function was called with
    Returns:
        Whether the files exist
    """
    feature_path = _get_feature_path(save_path)
    target_path = _get_target_path(save_path)

    return os.path.exists(feature_path) and os.path.exists(target_path)


def _get_feature_path(save_path):
    if save_path.endswith(".npy"):
        save_path = save_path[:-4]
    feature_path = f"{save_path}_features.npy"

    return feature_path


def _get_target_path(save_path):
    if save_path.endswith(".npy"):
        save_path = save_path[:-4]
    target_path = f"{save_path}_targets.npy"

    return target_path


def load_raw(
    file_paths: Dict[int, List[str]],
    window_size: int,
    columns: List[int],
    skip_rows: int = 0,
) -> Dict[int, np.ndarray]:
    features = {}
    for run_idx, run_files in tqdm(file_paths.items(), desc="Runs"):
        run_features = np.empty((len(run_files), window_size, len(columns)))
        for i, file_path in enumerate(tqdm(run_files, desc="Files", leave=False)):
            run_features[i] = _load_raw_features(file_path, skip_rows, columns)
        features[run_idx] = run_features

    return features


def _load_raw_features(
    file_path: str, skip_rows: int, columns: List[int]
) -> np.ndarray:
    try:
        features = _load_raw_unsafe(file_path, skip_rows)
    except ValueError:
        _replace_delimiters(file_path)
        features = _load_raw_unsafe(file_path, skip_rows)
    features = features[:, columns]

    return features


def _load_raw_unsafe(file_path: str, skip_rows: int) -> np.ndarray:
    return np.loadtxt(file_path, skiprows=skip_rows, delimiter=",")


def _replace_delimiters(file_path: str) -> None:
    with open(file_path, mode="r+t") as f:
        content = f.read()
        f.seek(0)
        content = content.replace(";", ",")
        f.write(content)
        f.truncate()



"""A module with functions for scaling RUL features."""
import copy
import pickle
from typing import List, Optional, Union, Tuple

import numpy as np
from sklearn import preprocessing as scalers  # type: ignore
from sklearn.base import BaseEstimator, TransformerMixin  # type: ignore

_Scaler = (
    scalers.StandardScaler,
    scalers.MinMaxScaler,
    scalers.MaxAbsScaler,
    scalers.RobustScaler,
)
Scaler = Union[
    scalers.StandardScaler,
    scalers.MinMaxScaler,
    scalers.MaxAbsScaler,
    scalers.RobustScaler,
]
"""
Supported scalers:

* [sklearn.preprocessing.StandardScaler][]
* [sklearn.preprocessing.MinMaxScaler][]
* [sklearn.preprocessing.MaxAbsScaler][]
* [sklearn.preprocessing.RobustScaler][]
"""


class OperationConditionAwareScaler(BaseEstimator, TransformerMixin):
    """This scaler is an ensemble of multiple base scalers, e.g. [
    sklearn.preprocessing.MinMaxScaler][]. It takes an additional operation condition
    array while fitting and transforming that controls which base scaler is used. The
    operation condition corresponding to a sample is compared against the boundaries
    defined during construction of the scaler. If the condition lies between the
    first set of boundaries, the first base scaler is used, and so forth.
    If any condition does not fall between any boundaries, an exception will be
    raised and the boundaries should be adjusted."""

    def __init__(
        self, base_scaler: Scaler, boundaries: List[Tuple[float, float]]
    ) -> None:
        """
        Create a new scaler aware of operation conditions.

        Each pair in `boundaries` represents the lower and upper value of an
        inclusive interval. For each interval a copy of the `base_scaler` is
        maintained. If an operation condition value falls inside an interval,
        the corresponding scaler is used. The boundaries have to be mutually exclusive.

        Args:
            base_scaler: The scaler that should be used for each condition.
            boundaries: The pairs that form the inclusive boundaries of each condition.
        """
        self.base_scalers = [copy.deepcopy(base_scaler) for _ in boundaries]
        self.boundaries = boundaries

        self._check_boundaries_mutually_exclusive()

    def _check_boundaries_mutually_exclusive(self):
        b = sorted(self.boundaries, key=lambda x: x[0])
        exclusive = all(up < low for (_, up), (low, _) in zip(b[:-1], b[1:]))
        if not exclusive:
            raise ValueError(
                "Boundaries are not mutually exclusive. Be aware that "
                "the boundaries are inclusive, i.e. lower <= value <= upper."
            )

    @property
    def n_features_in_(self):
        """Number of expected input features."""
        return self.base_scalers[0].n_features_in_

    def partial_fit(
        self, features: np.ndarray, operation_conditions: np.ndarray
    ) -> "OperationConditionAwareScaler":
        """
        Fit the base scalers partially.

        This function calls `partial_fit` on each of the base scalers with the
        samples that fall into the corresponding condition boundaries. If any sample
        does not fall into one of the boundaries, an exception is raised.

        Args:
            features: The feature array to be scaled.
            operation_conditions: The condition values compared against the boundaries.
        Returns:
            The partially fitted scaler.
        """
        total = 0
        for i, (lower, upper) in enumerate(self.boundaries):
            idx = self._between(operation_conditions, lower, upper)
            if num_elem := np.sum(idx):  # guard against empty array
                self.base_scalers[i].partial_fit(features[idx])
                total += num_elem
        self._check_all_transformed(features, total, "fitted")

        return self

    def transform(
        self, features: np.ndarray, operation_conditions: np.ndarray
    ) -> np.ndarray:
        """
        Scale the features with the appropriate condition aware scaler.

        This function calls `transform` on each of the base scalers for the
        samples that fall into the corresponding condition boundaries. If any sample
        does not fall into one of the boundaries, an exception is raised.

        Args:
            features: The features to be scaled.
            operation_conditions: The condition values compared against the boundaries.
        Returns:
            The scaled features.
        """
        scaled = np.empty_like(features)
        total = 0
        for i, (lower, upper) in enumerate(self.boundaries):
            idx = self._between(operation_conditions, lower, upper)
            if num_elem := np.sum(idx):  # guard against empty array
                scaled[idx] = self.base_scalers[i].transform(features[idx])
                total += num_elem
        self._check_all_transformed(features, total, "scaled")

        return scaled

    def _check_all_transformed(self, features, total, activity):
        """Guard against unknown conditions"""
        if diff := (len(features) - total):
            raise RuntimeError(
                f"{diff} samples had an unknown condition and could not be {activity}."
                "Please adjust the boundaries."
            )

    def _between(self, inputs: np.ndarray, lower: float, upper: float) -> np.ndarray:
        """Inclusive between."""
        return (lower <= inputs) & (inputs <= upper)


def fit_scaler(
    features: List[np.ndarray],
    scaler: Optional[Union[Scaler, OperationConditionAwareScaler]] = None,
    operation_conditions: Optional[List[np.ndarray]] = None,
) -> Union[Scaler, OperationConditionAwareScaler]:
    """
    Fit a given scaler to the RUL features. If the scaler is omitted,
    a StandardScaler will be created.

    If the scaler is an [OperationConditionAwareScaler][
    rul_datasets.reader.scaling.OperationConditionAwareScaler] and
    `operation_conditions` are passed, the scaler will be fit aware of operation
    conditions.

    The scaler assumes that the last axis of the features are the channels. Only
    scalers unaware of operation conditions can be fit with windowed data.

    Args:
        features: The RUL features.
        scaler: The scaler to be fit. Defaults to a StandardScaler.
        operation_conditions: The operation conditions for condition aware scaling.
    Returns:
        The fitted scaler.
    """
    scaler = scaler or scalers.StandardScaler()
    if isinstance(scaler, Scaler.__args__):  # type: ignore[attr-defined]
        scaler = _fit_scaler_naive(features, scaler)
    elif operation_conditions is not None and isinstance(
        scaler, OperationConditionAwareScaler
    ):
        scaler = _fit_scaler_operation_condition_aware(
            features, scaler, operation_conditions
        )
    else:
        raise ValueError(
            "Unsupported combination of scaler type and operation conditions."
        )

    return scaler


def _fit_scaler_naive(features: List[np.ndarray], scaler: Scaler) -> Scaler:
    for run in features:
        run = run.reshape(-1, run.shape[-1])
        scaler.partial_fit(run)

    return scaler


def _fit_scaler_operation_condition_aware(
    features: List[np.ndarray],
    scaler: OperationConditionAwareScaler,
    operation_conditions: List[np.ndarray],
) -> OperationConditionAwareScaler:
    assert len(features[0].shape) == 2, "Condition aware scaling can't fit window data"
    for run, cond in zip(features, operation_conditions):
        scaler.partial_fit(run, cond)

    return scaler


def save_scaler(scaler: Scaler, save_path: str) -> None:
    """
    Save a scaler to disk.

    Args:
        scaler: The scaler to be saved.
        save_path: The path to save the scaler to.
    """
    with open(save_path, mode="wb") as f:
        pickle.dump(scaler, f)


def load_scaler(save_path: str) -> Scaler:
    """
    Load a scaler from disk.

    Args:
        save_path: The path the scaler was saved to.
    Returns:
        The loaded scaler.
    """
    with open(save_path, mode="rb") as f:
        scaler = pickle.load(f)

    return scaler


def scale_features(
    features: List[np.ndarray],
    scaler: Union[Scaler, OperationConditionAwareScaler],
    operation_conditions: Optional[List[np.ndarray]] = None,
) -> List[np.ndarray]:
    """
    Scaler the RUL features with a given scaler.

    The features can have a shape of `[num_time_steps, channels]` or `[num_windows,
    window_size, channels]`. The scaler needs to work on the channel dimension. If it
    was not fit with the right number of channels, a `ValueError` is thrown.

    If the scaler is operation condition aware, the `operation_conditions` argument
    needs to be passed. Windowed data cannot be fit this way.

    Args:
        features: The RUL features to be scaled.
        scaler: The already fitted scaler.
        operation_conditions: The operation conditions for condition aware scaling.
    Returns:
        The scaled features.
    """
    if operation_conditions is None:
        features = _scale_features_naive(features, scaler)
    else:
        features = _scale_features_condition_aware(
            features, scaler, operation_conditions
        )

    return features


def _scale_features_naive(
    features: List[np.ndarray], scaler: Scaler
) -> List[np.ndarray]:
    features = copy.copy(features)
    for i, run in enumerate(features):
        _check_channels(run, scaler)
        if len(run.shape) == 3:
            features[i] = _scale_windowed_features(run, scaler)
        else:
            features[i] = scaler.transform(run)

    return features


def _scale_features_condition_aware(
    features: List[np.ndarray],
    scaler: OperationConditionAwareScaler,
    operation_conditions: List[np.ndarray],
) -> List[np.ndarray]:
    assert len(features[0].shape) == 2, "No condition aware scaling for window data"
    features = copy.copy(features)
    for i, (run, cond) in enumerate(zip(features, operation_conditions)):
        _check_channels(run, scaler)
        features[i] = scaler.transform(run, cond)

    return features


def _check_channels(
    run: np.ndarray, scaler: Union[Scaler, OperationConditionAwareScaler]
) -> None:
    if not run.shape[-1] == scaler.n_features_in_:
        raise ValueError(
            f"The scaler was fit on {scaler.n_features_in_} "
            f"channels but the features have {run.shape[1]} channels."
        )


def _scale_windowed_features(features: np.ndarray, scaler: Scaler) -> np.ndarray:
    num_channels = features.shape[2]
    window_size = features.shape[1]
    features = features.reshape(-1, num_channels)
    features = scaler.transform(features)
    features = features.reshape(-1, window_size, num_channels)

    return features



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rul_datasets', 'rul_datasets.reader']

package_data = \
{'': ['*']}

install_requires = \
['pytorch-lightning>=1.4.4,<2.0.0',
 'scikit-learn>=1.0.0,<2.0.0',
 'torch>=1.9.0,<2.0.0',
 'tqdm>=4.62.2,<5.0.0']

setup_kwargs = {
    'name': 'rul-datasets',
    'version': '0.11.0',
    'description': 'A collection of datasets for RUL estimation as Lightning Data Modules.',
    'long_description': '# RUL Datasets\n\n[![Master](https://github.com/tilman151/rul-datasets/actions/workflows/on_push.yaml/badge.svg)](https://github.com/tilman151/rul-datasets/actions/workflows/on_push.yaml)\n[![Release](https://github.com/tilman151/rul-datasets/actions/workflows/on_release.yaml/badge.svg)](https://github.com/tilman151/rul-datasets/actions/workflows/on_release.yaml)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nThis library contains a collection of common benchmark datasets for **remaining useful lifetime (RUL)** estimation.\nThey are provided as [LightningDataModules](https://pytorch-lightning.readthedocs.io/en/stable/api/pytorch_lightning.core.LightningDataModule.html#pytorch_lightning.core.LightningDataModule) to be readily used in [PyTorch Lightning](https://pytorch-lightning.readthedocs.io/en/latest/).\n\nCurrently, four datasets are supported:\n\n* **C-MAPSS** Turbofan Degradation Dataset\n* **FEMTO** (PRONOSTIA) Bearing Dataset\n* **XJTU-SY** Bearing Dataset\n* **Dummy** A tiny, simple dataset for debugging\n\nAll datasets share the same API, so they can be used as drop-in replacements for each other.\nThat means, if an experiment can be run with one of the datasets, it can be run with all of them.\nNo code changes needed.\n\nAside from the basic ones, this library contains data modules for advanced experiments concerning **transfer learning**, **unsupervised domain adaption** and **semi-supervised learning**.\nThese data modules are designed as **higher-order data modules**.\nThis means they take one or more of the basic data modules as inputs and adjust them to the desired use case.\n\n## Installation\n\nThe library is pip-installable. Simply type:\n\n```shell\npip install rul-datasets\n```\n\n## Contribution\n\nContributions are always welcome. Whether you want to fix a bug, add a feature or a new dataset, just open an issue and a PR.',
    'author': 'Krokotsch, Tilman',
    'author_email': 'tilman.krokotsch@tu-berlin.de',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://krokotsch.eu/rul-datasets',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""The XJTU-SY Bearing dataset is a collection of run-to-failure experiments on
bearings. Three different operation conditions were used, resulting in three
sub-datasets. Each sub-dataset contains five runs without an official training/test
split. """

import os.path
import tempfile
import zipfile
from typing import Tuple, List, Union, Dict, Optional

import numpy as np
from sklearn import preprocessing as scalers  # type: ignore

from rul_datasets import utils
from rul_datasets.reader import saving, scaling
from rul_datasets.reader.abstract import AbstractReader
from rul_datasets.reader.data_root import get_data_root

XJTU_SY_URL = "https://kr0k0tsch.de/rul-datasets/XJTU-SY.zip"


class XjtuSyReader(AbstractReader):
    """
    This reader represents the XJTU-SY Bearing dataset. Each of its three
    sub-datasets contains five runs. By default, the reader assigns the first two to
    the development, the third to the validation and the remaining two to the test
    split. This run to split assignment can be overridden by setting `run_split_dist`.

    The features contain windows with two channels of acceleration data which are
    standardized to zero mean and one standard deviation. The scaler is fitted on the
    development data.

    Examples:
        Default splits:
        >>> import rul_datasets
        >>> fd1 = rul_datasets.reader.XjtuSyReader(fd=1)
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> features[0].shape
        (123, 32768, 2)

        Custom splits:
        >>> import rul_datasets
        >>> splits = {"dev": [5], "val": [4], "test": [3]}
        >>> fd1 = rul_datasets.reader.XjtuSyReader(fd=1, run_split_dist=splits)
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> features[0].shape
        (52, 32768, 2)

        Set first-time-to-predict:
        >>> import rul_datasets
        >>> fttp = [10, 20, 30, 40, 50]
        >>> fd1 = rul_datasets.reader.XjtuSyReader(fd=1, first_time_to_predict=fttp)
        >>> fd1.prepare_data()
        >>> features, labels = fd1.load_split("dev")
        >>> labels[0][:15]
        array([113., 113., 113., 113., 113., 113., 113., 113., 113., 113., 113.,
               112., 111., 110., 109.])
    """

    _XJTU_SY_ROOT: str = os.path.join(get_data_root(), "XJTU-SY")
    _NUM_TRAIN_RUNS: Dict[int, int] = {1: 5, 2: 5, 3: 5}

    def __init__(
        self,
        fd: int,
        window_size: Optional[int] = None,
        max_rul: Optional[int] = None,
        percent_broken: Optional[float] = None,
        percent_fail_runs: Optional[Union[float, List[int]]] = None,
        truncate_val: bool = False,
        run_split_dist: Optional[Dict[str, List[int]]] = None,
        first_time_to_predict: Optional[List[int]] = None,
        norm_rul: bool = False,
        truncate_degraded_only: bool = False,
    ) -> None:
        """
        Create a new XJTU-SY reader for one of the sub-datasets. By default, the RUL
        values are not capped. The default window size is 32768.

        Use `first_time_to_predict` to set an individual RUL inflection point for
        each run. It should be a list with an integer index for each run. The index
        is the time step after which RUL declines. Before the time step it stays
        constant. The `norm_rul` argument can then be used to scale the RUL of each
        run between zero and one.

        For more information about using readers, refer to the [reader]
        [rul_datasets.reader] module page.

        Args:
            fd: Index of the selected sub-dataset
            window_size: Size of the sliding window. Defaults to 32768.
            max_rul: Maximum RUL value of targets.
            percent_broken: The maximum relative degradation per time series.
            percent_fail_runs: The percentage or index list of available time series.
            truncate_val: Truncate the validation data with `percent_broken`, too.
            run_split_dist: Dictionary that assigns each run idx to each split.
            first_time_to_predict: The time step for each time series before which RUL
                                   is constant.
            norm_rul: Normalize RUL between zero and one.
            truncate_degraded_only: Only truncate the degraded part of the data
                                    (< max RUL).
        """
        super().__init__(
            fd,
            window_size,
            max_rul,
            percent_broken,
            percent_fail_runs,
            truncate_val,
            truncate_degraded_only,
        )

        if (first_time_to_predict is not None) and (max_rul is not None):
            raise ValueError(
                "XjtuSyReader cannot use 'first_time_to_predict' "
                "and 'max_rul' in conjunction."
            )

        self.first_time_to_predict = first_time_to_predict
        self.norm_rul = norm_rul

        self._preparator = XjtuSyPreparator(self.fd, self._XJTU_SY_ROOT, run_split_dist)

    @property
    def dataset_name(self) -> str:
        return "xjtu-sy"

    @property
    def fds(self) -> List[int]:
        """Indices of available sub-datasets."""
        return list(self._NUM_TRAIN_RUNS)

    def prepare_data(self) -> None:
        """
        Prepare the XJTU-SY dataset. This function needs to be called before using the
        dataset and each custom split for the first time.

        The dataset is downloaded from a custom mirror and extracted into the data
        root directory. The whole dataset is converted com CSV files to NPY files to
        speed up loading it from disk. Afterwards, a scaler is fit on the development
        features. Previously completed steps are skipped.
        """
        if not os.path.exists(self._XJTU_SY_ROOT):
            _download_xjtu_sy(get_data_root())
        self._preparator.prepare_split()

    def load_complete_split(
        self, split: str, alias: str
    ) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        features, targets = self._preparator.load_runs(split)
        features = [f[:, -self.window_size :, :] for f in features]  # crop to window
        features = scaling.scale_features(features, self._preparator.load_scaler())
        if self.max_rul is not None:
            targets = [np.minimum(t, self.max_rul) for t in targets]
        elif self.first_time_to_predict is not None:
            targets = self._clip_first_time_to_predict(targets, split)

        if self.norm_rul:
            targets = [t / np.max(t) for t in targets]

        return features, targets

    def _clip_first_time_to_predict(self, targets, split):
        fttp = [
            self.first_time_to_predict[i - 1]
            for i in self._preparator.run_split_dist[split]
        ]
        targets = [np.minimum(t, len(t) - fttp) for t, fttp in zip(targets, fttp)]

        return targets

    def default_window_size(self, fd: int) -> int:
        return XjtuSyPreparator.DEFAULT_WINDOW_SIZE


class XjtuSyPreparator:
    DEFAULT_WINDOW_SIZE: int = 32768
    _FD_FOLDERS: Dict[int, str] = {1: "35Hz12kN", 2: "37.5Hz11kN", 3: "40Hz10kN"}
    _DEFAULT_RUN_SPLIT_DIST: Dict[str, List[int]] = {
        "dev": [1, 2],
        "val": [3],
        "test": [4, 5],
    }

    def __init__(
        self,
        fd: int,
        data_root: str,
        run_split_dist: Optional[Dict[str, List[int]]] = None,
    ) -> None:
        self.fd = fd
        self.data_root = data_root
        self.run_split_dist = run_split_dist or self._DEFAULT_RUN_SPLIT_DIST

    def prepare_split(self, split: Optional[str] = None) -> None:
        run_file_path = self._get_run_file_path(1)
        if not saving.exists(run_file_path):
            runs = self._load_raw_runs()
            runs = self._sort_runs(runs)
            self._save_efficient(runs)
        if not os.path.exists(self._get_scaler_path()):
            features, _ = self.load_runs("dev")
            scaler = scaling.fit_scaler(features)
            scaling.save_scaler(scaler, self._get_scaler_path())

    def load_runs(self, split: str) -> Tuple[List[np.ndarray], List[np.ndarray]]:
        self._validate_split(split)
        run_idx = self.run_split_dist[split]
        run_paths = [self._get_run_file_path(idx) for idx in run_idx]
        features, targets = saving.load_multiple(run_paths)

        return features, targets

    def load_scaler(self) -> scalers.StandardScaler:
        return scaling.load_scaler(self._get_scaler_path())

    def _load_raw_runs(self) -> Dict[int, Tuple[np.ndarray, np.ndarray]]:
        file_paths = self._get_csv_file_paths()
        features = saving.load_raw(
            file_paths, self.DEFAULT_WINDOW_SIZE, columns=[0, 1], skip_rows=1
        )
        targets = utils.get_targets_from_file_paths(
            file_paths, self._timestep_from_file_path
        )
        runs = {idx: (features[idx], targets[idx]) for idx in features}

        return runs

    def _sort_runs(
        self, runs: Dict[int, Tuple[np.ndarray, np.ndarray]]
    ) -> Dict[int, Tuple[np.ndarray, np.ndarray]]:
        sort_idx = {run_idx: np.argsort(t)[::-1] for run_idx, (_, t) in runs.items()}
        runs = {
            run_idx: (f[sort_idx[run_idx]], t[sort_idx[run_idx]])
            for run_idx, (f, t) in runs.items()
        }

        return runs

    def _get_csv_file_paths(self) -> Dict[int, List[str]]:
        fd_folder_path = self._get_fd_folder_path()
        file_paths = {}
        run_folders = self._get_run_folders(fd_folder_path)
        for run_idx, run_folder in run_folders.items():
            run_path = os.path.join(fd_folder_path, run_folder)
            feature_files = utils.get_files_in_path(run_path)
            file_paths[run_idx] = feature_files

        return file_paths

    def _get_run_folders(self, split_path: str) -> Dict[int, str]:
        all_folders = sorted(os.listdir(split_path))
        run_folders = {
            int(f[-1]): f
            for f in all_folders
            if os.path.isdir(os.path.join(split_path, f))
        }

        return run_folders

    @staticmethod
    def _timestep_from_file_path(file_path: str) -> int:
        file_name = os.path.basename(file_path)
        time_step = int(file_name.replace(".csv", ""))

        return time_step

    def _save_efficient(self, runs) -> None:
        for run_idx, (features, targets) in runs.items():
            saving.save(self._get_run_file_path(run_idx), features, targets)

    def _validate_split(self, split: str) -> None:
        if split not in ["dev", "val", "test"]:
            raise ValueError(
                "XJTU-SY provides a dev, val and test split, "
                f"but you provided '{split}'."
            )

    def _get_scaler_path(self) -> str:
        file_name = f"scaler_{self.run_split_dist['dev']}.pkl"
        file_path = os.path.join(self._get_fd_folder_path(), file_name)

        return file_path

    def _get_run_file_path(self, run_idx: int) -> str:
        return os.path.join(self._get_fd_folder_path(), f"run_{run_idx}.npy")

    def _get_fd_folder_path(self) -> str:
        return os.path.join(self.data_root, self._FD_FOLDERS[self.fd])


def _download_xjtu_sy(data_root: str) -> None:
    with tempfile.TemporaryDirectory() as tmp_path:
        print("Download XJTU-SY dataset")
        download_path = os.path.join(tmp_path, "XJTU-SY.zip")
        utils.download_file(XJTU_SY_URL, download_path)
        print("Extract XJTU-SY dataset")
        with zipfile.ZipFile(download_path, mode="r") as f:
            f.extractall(data_root)



__version__ = "0.11.0"

import warnings

from .adaption import (
    DomainAdaptionDataModule,
    PretrainingAdaptionDataModule,
    LatentAlignDataModule,
)
from .baseline import BaselineDataModule, PretrainingBaselineDataModule
from .core import RulDataModule
from .reader import CmapssReader, FemtoReader, XjtuSyReader
from .reader.data_root import get_data_root, set_data_root
from .ssl import SemiSupervisedDataModule

warnings.filterwarnings(
    "ignore", ".*Consider increasing the value of the `num_workers` argument*"
)
